import pathlib
import typing
import sys
import natsort
import threading
from dataclasses import dataclass
from imgui_bundle import hello_imgui, imgui, icons_fontawesome_6 as ifa6, imspinner

from .. import file_actions, file_action_provider as fap, platform, utils
from . import msg_box, utils as gui_utils


DRIVE_ICON = ifa6.ICON_FA_HARD_DRIVE
SERVER_ICON = ifa6.ICON_FA_SERVER
DIR_ICON = ifa6.ICON_FA_FOLDER
FILE_ICON = ifa6.ICON_FA_FILE


@dataclass
class DirEntryWithCache(file_actions.DirEntry):
    # display fields
    display_name: str = ''
    ctime_str: str = None
    mtime_str: str = None
    size_str: str = None

    def __init__(self, item: file_actions.DirEntry):
        super().__init__(item.name, item.is_dir, item.full_path, item.ctime, item.mtime, item.size, item.mime_type, item.extra)

        # prep display strings
        if self.mime_type and self.mime_type.startswith('file_action/drive'):
            icon = DRIVE_ICON
        elif self.mime_type=='file_action/net_name':
            icon = SERVER_ICON
        elif self.is_dir:
            icon = DIR_ICON
        else:
            icon = FILE_ICON
        self.display_name   = icon + "  " + self.name

        self.ctime_str      = self.ctime.strftime("%Y-%m-%d %H:%M:%S") if self.ctime else None
        self.mtime_str      = self.mtime.strftime("%Y-%m-%d %H:%M:%S") if self.mtime else None

        # size
        if not self.is_dir or (self.mime_type and self.mime_type.startswith('file_action/drive')):
            i = 0
            units = ['B', 'KiB', 'MiB', 'GiB', 'TiB']
            size = self.size
            while size>1024:
                i+=1
                size /= 1024
            if i==0:
                self.size_str = f'{size:.0f} {units[i]}'
            else:
                self.size_str = f'{size:.1f} {units[i]}'

class FilePicker:
    default_flags: int = (
        imgui.WindowFlags_.no_collapse |
        imgui.WindowFlags_.no_saved_settings
    )

    def __init__(self, title="File picker", dir_picker=False, start_dir: str | pathlib.Path = None, callback: typing.Callable = None, allow_multiple = True, file_action_provider: fap.FileActionProvider = None, custom_popup_flags=0):
        self.title = title
        self.elapsed = 0.0
        self.callback = callback
        if file_action_provider:
            self.file_action_provider = file_action_provider
        else:
            self.file_action_provider = fap.FileActionProvider(self._listing_done, self._action_done)

        self._listing_cache: dict[str|pathlib.Path, dict[int,DirEntryWithCache]] = {}
        self.popup_stack = []
        self.dialog_provider = DialogProvider(self, self._launch_action)
        self.disable_keyboard_navigation = False

        self.items: dict[int, DirEntryWithCache] = {}
        self.selected: dict[int, bool] = {}
        self.items_lock: threading.Lock = threading.Lock()  # for self.items and self.selected
        self.msg: str = None
        self.filter_box_text = ''
        self.require_sort = False
        self.sorted_items: list[int] = []
        self.last_clicked_id: int = None

        self.allow_multiple = allow_multiple
        self._is_dir_picker = False
        self.predicate_selectable: typing.Callable[[int], bool] = None
        self.set_is_dir_picker(dir_picker)
        self._show_only_dirs = False
        self.predicate_showable: typing.Callable[[int], bool] = None
        self.set_show_only_dirs(self._is_dir_picker)   # by default, a dir picker only shows dirs

        self.accent_color: tuple[float] = None
        self.bg_color: tuple[float] = None

        self.loc: pathlib.Path = None
        self.refreshing = False
        self.new_loc = False
        self.select_when_loaded: list[pathlib.Path] = None
        self.history: list[str|pathlib.Path] = []
        self.history_loc = -1
        self.path_bar_popup: dict[str,typing.Any] = {}
        self.default_flags = custom_popup_flags or FilePicker.default_flags
        self.platform_is_windows = sys.platform.startswith("win")

        self.goto(start_dir or '.')
        self._request_listing('root')   # request root listing so we have the drive names

    def set_is_dir_picker(self, is_dir_picker):
        self._is_dir_picker = is_dir_picker
        if self._is_dir_picker:
            self.predicate_selectable = lambda iid: self.items[iid].is_dir
        else:
            self.predicate_selectable = None

    def set_show_only_dirs(self, do_show_only_dirs):
        self._show_only_dirs = do_show_only_dirs
        if self._show_only_dirs:
            self.predicate_showable = lambda iid: self.items[iid].is_dir
        else:
            self.predicate_showable = None

    def set_draw_parameters(self, accent_color: tuple[float] = None, bg_color: tuple[float] = None):
        self.accent_color = accent_color
        self.bg_color = bg_color

    def _is_root(self, path: str|pathlib.Path):
        if platform.os==platform.Os.Windows and isinstance(path, str):
            if path.casefold()==self._get_path_display_name('root').casefold():
                path = 'root'
            return path=='root'
        elif platform.os!=platform.Os.Windows:
            return path=='root' or (pa:=pathlib.Path(path).absolute())==pathlib.Path(pa.anchor)
        return False


    def set_dir(self, paths: str|pathlib.Path|list[str|pathlib.Path]):
        if not isinstance(paths,list):
            paths = [paths]
        paths = [pathlib.Path(p) for p in paths]

        if len(paths)==1 and paths[0].is_dir():
            self.goto(paths[0])
        else:
            self.select_when_loaded = paths
            self.goto(paths[0].parent)

    def goto(self, path: str | pathlib.Path, add_history=True):
        is_root = self._is_root(path)

        if not is_root:
            if platform.os==platform.Os.Windows and (comp := file_actions.get_net_computer(path)):
                # ensure loc has the format //SERVER/, which is what pathlib understands
                # but keep it as a string, these things don't round-trip in pathlib
                path = f'//{comp}/'
            else:
                path = pathlib.Path(path).expanduser().resolve()
                if path.is_file():
                    path = path.parent
                if path is None:
                    path = pathlib.Path('.')
                if not str(path).startswith('\\\\') or str(path).startswith('//'):
                    # don't call resolve on network paths
                    path = path.resolve()
        else:
            path = 'root'

        if path!=self.loc:
            self.loc = path
            self.new_loc = True
            if add_history:
                if self.history_loc>=0:
                    # remove any history after current, as we're appending a new location
                    del self.history[self.history_loc+1:]
                self.history.append(self.loc)
                self.history_loc += 1
            # changing location clears selection
            with self.items_lock:
                utils.set_all(self.selected, False)
            # changing location clear filter box
            self.filter_box_text = ''
            # load from cache if available
            if self.loc in self._listing_cache:
                self._update_listing(self.loc, True)
            # load new directory
            self.refresh()

    def refresh(self):
        # launch refresh
        self.refreshing = True
        self._request_listing(self.loc)

    def _request_listing(self, path: str|pathlib.Path):
        self.file_action_provider.get_listing(path)

    def _listing_done(self, path: str|pathlib.Path, items: list[file_actions.DirEntry]|Exception):
        # deal with cache
        if not isinstance(items, Exception):
            items = {i:DirEntryWithCache(item) for i,item in enumerate(natsort.os_sorted(items, key=lambda i: i.full_path))}
        self._listing_cache[path] = items

        if str(path)==str(self.loc):
            # also load all parent paths if not in cache already, so path bar
            # drop downs work
            loc = self.loc
            while loc:
                if loc not in self._listing_cache:
                    self._request_listing(loc)
                loc = self._get_parent(loc)

            # and update the shown listing
            self._update_listing(path, False)

    def _update_listing(self, path: str|pathlib.Path, from_cache: bool):
        previously_selected = []
        with self.items_lock:
            if self.select_when_loaded:
                previously_selected = self.select_when_loaded
                self.select_when_loaded = None
            elif not self.new_loc:
                previously_selected = [self.items[iid].full_path for iid in self.items if iid in self.selected and self.selected[iid]]
            self.items.clear()
            self.selected.clear()
            self.msg = None
            items = self._listing_cache[path]
            if isinstance(items, Exception):
                self.msg = f"Cannot open this folder!\n:{items}"
            else:
                self.items = items.copy()
                self.selected = {k:False for k in self.items}

        # if refreshed the same directory, restore old selection
        self._select_paths(previously_selected)

        self.require_sort = True
        self.new_loc = False
        if not from_cache:
            self.refreshing = False
            self.elapsed = 0.0

    def _launch_action(self, action: str, path: str|pathlib.Path, path2: str|pathlib.Path = None):
        match action:
            case 'make_dir':
                self.file_action_provider.make_dir(path)
            case 'rename_path':
                self.file_action_provider.rename_path(path, path2)

    def _action_done(self, path: pathlib.Path, action: str, result: None|pathlib.Path|Exception):
        if isinstance(result, Exception):
            match action:
                case 'make_dir':
                    action_lbl = 'making the folder'
                case 'rename_path':
                    action_lbl = 'renaming the folder or file'
            gui_utils.push_popup(self, msg_box.msgbox, "Action error", f'Something went wrong {action_lbl} {path}":\n{result}', msg_box.MsgBox.error)

        # trigger refresh of parent path where actions occurred
        self._request_listing(path.parent)
        # if there is a result path and it has a different parent than the action path, refresh that one too
        if isinstance(result, pathlib.Path) and result.parent!=path.parent:
            self._request_listing(result.parent)


    def _select_paths(self, paths: list[pathlib.Path]):
        got_one = False
        with self.items_lock:
            for path in paths:
                for iid in self.items:
                    entry = self.items[iid]
                    if entry.full_path==path and (not self.predicate_selectable or self.predicate_selectable(iid)):
                        self.selected[iid] = True
                        got_one = True
                        break
                if not self.allow_multiple and got_one:
                    break

    def _dir_picker_is_current_dir_selected(self):
        with self.items_lock:
            any_selected = any((self.selected[iid] for iid in self.selected))
        return self._is_dir_picker and not any_selected and not self._is_root(self.loc)

    def _get_parent(self, path: str | pathlib.Path):
        if isinstance(path,str) and path=='root':
            return None
        if not isinstance(path,pathlib.Path):
            path = pathlib.Path(path)
        parent = path.parent
        # on Posix we check if the parent is the fs anchor, that means the parent is root
        # on Windows, there is a level above the fs anchor, so we check if the current path (and not its parent) is the anchor
        to_check = path if platform.os==platform.Os.Windows else parent
        if to_check==pathlib.Path(path.anchor):
            if isinstance(path,pathlib.PureWindowsPath):
                if platform.os==platform.Os.Windows and (net_comps := file_actions.split_network_path(path)):
                    if len(net_comps)==1:
                        # network computer, one higher is list of drives and
                        # visible network computers, i.e., root
                        parent = 'root'
                    else:
                        parent = pathlib.Path(f'//{"/".join(net_comps[:-1])}/')
                else:
                    # if a normal local path, then this means we're in a drive root
                    parent = 'root'
            else:
                # non-Windows logic: simply root
                parent = 'root'
        return parent

    def _get_path_display_name(self, path: str | pathlib.Path, for_edit=False):
        path_str = str(path)
        if path_str=='root' or self._is_root(path):
            loc_str = self.file_action_provider.local_name
        else:
            if platform.os==platform.Os.Windows:
                if (comp := file_actions.get_net_computer(path_str)):
                    # pathlib.Path's str() doesn't do the right thing here, render it ourselves
                    loc_str = f'\\\\{comp}'
                else:
                    if not for_edit and isinstance(path,pathlib.Path) and self._get_parent(path)=='root' and 'root' in self._listing_cache and isinstance(self._listing_cache['root'],dict):
                        # this is a drive root, lookup name of this drive
                        # except when this is meant to be a string to be edited, then we
                        # just want the raw drive location string
                        loc_str = None
                        for _,l in self._listing_cache['root'].items():
                            if l.full_path==path:
                                loc_str = l.name
                        if loc_str is None:
                            loc_str = path.drive
                    else:
                        loc_str = path_str
            else:
                loc_str = path.name if not for_edit else path_str
        return loc_str

    def _get_path_leaf_display_name(self, path: str | pathlib.Path):
        if isinstance(path, pathlib.Path) and self._get_parent(path)!='root':
            if platform.os==platform.Os.Windows and (net_comps := file_actions.split_network_path(path)):
                if len(net_comps)==1:
                    disp_name = f'\\\\{net_comps[0]}'
                else:
                    disp_name = net_comps[-1]
            else:
                disp_name = path.name
                if not disp_name:
                    # disk root
                    disp_name = str(path)
        else:
            # special string
            disp_name = self._get_path_display_name(path)
        return disp_name


    def draw(self):
        cancelled = closed = False

        imgui.begin_child('##filepicker')
        self.draw_top_bar()
        closed = self.draw_listing(leave_space_for_bottom_bar=True)
        cancelled, closed2 = self.draw_bottom_bar()
        imgui.end_child()

        closed = closed or closed2
        return cancelled, closed

    def draw_top_bar(self):
        self.elapsed += imgui.get_io().delta_time

        enable_keyboard_nav = not self.popup_stack and not self.disable_keyboard_navigation and not imgui.get_io().want_text_input   # no keyboard navigation in this GUI if a popup is open or key input taken elsewhere
        backspace_released  = imgui.is_key_pressed(imgui.Key.backspace)
        shift_down          = imgui.is_key_down(imgui.Key.im_gui_mod_shift)

        imgui.begin_group()
        # History back button
        disabled = self.history_loc<=0
        if disabled:
            imgui.begin_disabled()
        if imgui.button(ifa6.ICON_FA_ARROW_LEFT) or (not disabled and enable_keyboard_nav and backspace_released and not shift_down):
            self.history_loc -= 1
            self.goto(self.history[self.history_loc], add_history=False)
        if self.history_loc>0: # don't just use disabled var as we may have just changed self.history_loc
            gui_utils.draw_hover_text(self._get_path_display_name(self.history[self.history_loc-1]), '')
            if imgui.begin_popup_context_item(f"##history_back_context"):
                for i in range(self.history_loc-1,-1,-1):
                    p = self.history[i]
                    if imgui.selectable(self._get_path_display_name(p), False)[0]:
                        self.history_loc = i
                        self.goto(p, add_history=False)
                imgui.end_popup()
        if disabled:
            imgui.end_disabled()
        # History forward button
        imgui.same_line(spacing=imgui.get_style().item_spacing.x/4)
        disabled = self.history_loc+1>=len(self.history)
        if disabled:
            imgui.begin_disabled()
        if imgui.button(ifa6.ICON_FA_ARROW_RIGHT) or (not disabled and enable_keyboard_nav and backspace_released and shift_down):
            self.history_loc += 1
            self.goto(self.history[self.history_loc], add_history=False)
        if self.history_loc+1<len(self.history): # don't just use disabled var as we may have just changed self.history_loc
            gui_utils.draw_hover_text(self._get_path_display_name(self.history[self.history_loc+1]), '')
            if imgui.begin_popup_context_item(f"##history_forward_context"):
                for i in range(self.history_loc+1,len(self.history)):
                    p = self.history[i]
                    if imgui.selectable(self._get_path_display_name(p), False)[0]:
                        self.history_loc = i
                        self.goto(p, add_history=False)
                imgui.end_popup()
        if disabled:
            imgui.end_disabled()
        # Up button
        imgui.same_line(spacing=imgui.get_style().item_spacing.x/4)
        parent = self._get_parent(self.loc)
        disabled = parent is None or self.loc==parent
        if disabled:
            imgui.begin_disabled()
        if imgui.button(ifa6.ICON_FA_ARROW_UP):
            self.goto(parent)
        if not disabled:
            gui_utils.draw_hover_text(self._get_path_display_name(parent), '')
        if disabled:
            imgui.end_disabled()
        # Refresh button
        imgui.same_line(spacing=imgui.get_style().item_spacing.x/2)
        if self.refreshing:
            button_text_size = imgui.calc_text_size(ifa6.ICON_FA_ARROW_ROTATE_RIGHT)
            button_size = (button_text_size.x+imgui.get_style().frame_padding.x*2, button_text_size.y+imgui.get_style().frame_padding.y*2)
            symbol_size = imgui.calc_text_size("x").y/2
            spinner_radii = [x/22*symbol_size for x in [22, 16, 10]]
            lw = 3.5/22*symbol_size
            spinner_diam = 2*spinner_radii[0]+lw
            offset_x = (button_size[0]-spinner_diam)/2
            cp = imgui.get_cursor_pos()
            imgui.set_cursor_pos_x(cp.x+offset_x)
            imspinner.spinner_ang_triple(f'loadingSpinner', *spinner_radii, lw, c1=imgui.get_style().color_(imgui.Col_.text_selected_bg), c2=imgui.get_style().color_(imgui.Col_.text), c3=imgui.get_style().color_(imgui.Col_.text_selected_bg))
            imgui.set_cursor_pos(cp)
            imgui.dummy(button_size)
            gui_utils.draw_hover_text(text='', hover_text='Refreshing...')
        else:
            if imgui.button(ifa6.ICON_FA_ARROW_ROTATE_RIGHT):
                self.refresh()
        # Location bar
        imgui.same_line(spacing=imgui.get_style().item_spacing.x/2)
        # determine size
        space = imgui.get_content_region_avail().x
        if space>250/.3*hello_imgui.dpi_window_size_factor():
            filt_space = 250*hello_imgui.dpi_window_size_factor()
        else:
            filt_space = int(space*.3)
        imgui.set_next_item_width(-filt_space)
        self.draw_path_bar()
        # filter box
        imgui.same_line(spacing=imgui.get_style().item_spacing.x/2)
        imgui.set_next_item_width(imgui.get_content_region_avail().x)
        _, value = imgui.input_text_with_hint('##filter_box', f'Filter {self._get_path_leaf_display_name(self.loc)}', self.filter_box_text)
        if value != self.filter_box_text:
            self.filter_box_text = value
            self.require_sort = True
        imgui.end_group()

    def draw_path_bar(self):
        # port of ImFileDialog's PathBar, edited and extended to work the way i like it
        win = imgui.internal.get_current_window()
        if win.skip_items:
            return

        iid = win.get_id('##path_bar')
        state = win.state_storage.get_int_ref(iid, 0)   # bit 1: button mode?, bit 2: hovered?, bit 3: set keyboard focus

        ctx = imgui.get_current_context()
        w = imgui.calc_item_width()
        ts = imgui.calc_text_size('x')
        pos = imgui.get_cursor_screen_pos()
        ui_pos = imgui.get_cursor_pos()
        bb = imgui.internal.ImRect(pos, (pos.x+w, pos.y+ts.y+2*imgui.get_style().frame_padding.y))

        if not state & 0b001:
            # buttons
            imgui.push_clip_rect(bb.min, bb.max, False)
            hovered = ctx.io.mouse_pos.x >= bb.min.x and ctx.io.mouse_pos.x <= bb.max.x and \
                      ctx.io.mouse_pos.y >= bb.min.y and ctx.io.mouse_pos.y <= bb.max.y
            clicked = hovered and imgui.is_mouse_released(imgui.MouseButton_.left)
            path_element_hc = False # are any other of the path boxes hovered or clicked?
            frame_col = imgui.get_color_u32(imgui.Col_.frame_bg_hovered if state & 0b010 else imgui.Col_.frame_bg)
            imgui.internal.render_frame(bb.min,bb.max,frame_col,True,imgui.get_style().frame_rounding)

            # get path components
            make_elem = lambda x, lbl: (lbl, x, imgui.calc_text_size(x).x)
            separator = '>'
            btn_list = []
            loc = self.loc
            while loc:
                btn_list.append(make_elem(self._get_path_leaf_display_name(loc), loc))
                btn_list.append(make_elem(separator, 'sep'))
                loc = self._get_parent(loc)
            del btn_list[-1]    # remove last separator
            btn_list.reverse()
            # simply show local machine as an icon, show name upon hover
            btn_list[0] = (btn_list[0][0], (ifa6.ICON_FA_DESKTOP, btn_list[0][1]), btn_list[0][2])

            # check if whole path fits, else shorten
            get_total_width = lambda x: sum([b[2] for b in x]) + len(btn_list)*2*imgui.get_style().frame_padding.x
            total_width = get_total_width(btn_list)
            btn_removed = []
            if total_width>w:
                ellipsis = make_elem('···', 'ellipsis')
                btn_list.insert(1, ellipsis)
                while total_width>w:
                    if len(btn_list)<=4:
                        # nothing more that can be removed
                        break
                    btn_removed.append(btn_list.pop(3))
                    del btn_list[3] # also remove separator that follows
                    total_width = get_total_width(btn_list)

            # draw buttons on the frame
            imgui.push_style_var(imgui.StyleVar_.item_spacing, (0, imgui.get_style().item_spacing.y))
            imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0)
            open_popup = False
            for i,b in enumerate(btn_list):
                id_str      = f'###path_comp_{i}'
                button_pos  = imgui.get_cursor_screen_pos()
                lbl = b[1]
                hover = None
                if isinstance(lbl, tuple):
                    hover = lbl[1]
                    lbl = lbl[0]
                if 'which' in self.path_bar_popup and \
                    self.path_bar_popup['which']==i-1 and \
                    imgui.is_popup_open('##dir_list_popup'):
                    lbl = 'v'
                clicked_button = False
                if lbl=='>':
                    clicked_button = imgui.arrow_button(id_str,imgui.Dir_.right)
                elif lbl=='v':
                    clicked_button = imgui.arrow_button(id_str,imgui.Dir_.down)
                else:
                    clicked_button = imgui.button(lbl+id_str)
                if clicked_button:
                    if isinstance(b[0],str):
                        if b[0]=='sep':
                            # path separator button, enqueue opening path selection dropdown
                            self.path_bar_popup['loc'] = self._get_parent(btn_list[i+1][0])
                            self.path_bar_popup['which'] = i-1
                            self.path_bar_popup['which_selected'] = btn_list[i+1][1]
                            self.path_bar_popup['pos'] = button_pos
                            open_popup = True
                        elif b[0]=='ellipsis':
                            # draw dropdown with removed paths
                            self.path_bar_popup['loc'] = 'ellipsis'
                            self.path_bar_popup['which'] = None
                            self.path_bar_popup['pos'] = button_pos
                            open_popup = True
                        else:
                            self.goto(b[0])
                    else:
                        self.goto(b[0])
                path_element_hc = path_element_hc or imgui.is_item_hovered() or imgui.is_item_clicked()
                if hover:
                    gui_utils.draw_hover_text(hover, '')
                imgui.same_line()
            imgui.pop_style_var(2)
            imgui.pop_clip_rect()

            do_open_popup = False
            if open_popup:
                # directory selector
                if (isinstance(self.path_bar_popup['loc'],str) and self.path_bar_popup['loc']=='ellipsis') or \
                    (self.path_bar_popup['loc'] in self._listing_cache and isinstance(self._listing_cache[self.path_bar_popup['loc']], dict)):
                    imgui.open_popup('##dir_list_popup')
                    do_open_popup = True
            if do_open_popup:
                # position popup: move button height down so it pops under the button
                self.path_bar_popup['pos'].y += imgui.calc_text_size('x').y+2*imgui.get_style().frame_padding.y+imgui.get_style().item_spacing.y
                imgui.set_next_window_pos(self.path_bar_popup['pos'])
            if imgui.begin_popup('##dir_list_popup'):
                key = self.path_bar_popup['loc']
                if key in self._listing_cache:
                    items = self._listing_cache[key]
                    items = [items[i] for i in items if items[i].is_dir]
                    paths         = [i.full_path for i in items]
                    display_names = [self._get_path_leaf_display_name(p) for p in paths]
                    idx = display_names.index(self.path_bar_popup['which_selected'])
                elif isinstance(key,str) and key=='ellipsis':
                    display_names = [b[1] for b in btn_removed]
                    paths = [b[0] for b in btn_removed]
                    display_names.reverse()
                    paths.reverse()
                    idx = -1
                changed, idx = imgui.list_box('##dir_list_popup_select',idx,display_names)
                if changed:
                    self.goto(paths[idx])
                    imgui.close_current_popup()
                imgui.end_popup()


            # click state
            if not path_element_hc and clicked:
                state |= 0b001
                state &= 0b011      # remove SetKeyboardFocus flag
            else:
                state &= 0b110
            # hover state
            if not path_element_hc and hovered and not clicked:
                state |= 0b010
            else:
                state &= 0b101
            # allocate space
            imgui.set_cursor_pos(ui_pos)
            imgui.internal.item_size(bb)
        else:
            # input box
            skip_active_check = False
            if not state & 0b100:
                skip_active_check = True
                imgui.set_keyboard_focus_here()
                if not imgui.is_mouse_clicked(imgui.MouseButton_.left):
                    state |= 0b100

            loc_str = self._get_path_display_name(self.loc, for_edit=True)
            confirmed, loc = imgui.input_text("##pathbox_input",loc_str,imgui.InputTextFlags_.enter_returns_true)
            if confirmed:
                self.goto(loc)
            if not skip_active_check and not imgui.is_item_active():
                state &= 0b010

        win.state_storage.set_int(iid, state)

    def draw_bottom_bar(self):
        cancelled = closed = False

        # Cancel button
        if imgui.button(ifa6.ICON_FA_CIRCLE_XMARK+" Cancel"):
            imgui.close_current_popup()
            cancelled = closed = True
        # Ok button
        imgui.same_line()
        with self.items_lock:
            selected = [self.items[iid] for iid in self.items if iid in self.selected and self.selected[iid]]
        num_selected = len(selected)
        dir_picker_selected = self._dir_picker_is_current_dir_selected()
        disable_ok = not num_selected and not dir_picker_selected or (self.refreshing and self.new_loc)
        if disable_ok:
            imgui.begin_disabled()
        if imgui.button(ifa6.ICON_FA_CHECK+" Ok"):
            imgui.close_current_popup()
            closed = True
        if disable_ok:
            imgui.end_disabled()
        # Selected text
        imgui.same_line()
        if dir_picker_selected:
            imgui.text(f"  Selected the current directory ({self._get_path_leaf_display_name(self.loc)})")
        elif num_selected==1:
            imgui.text(f"  Selected {num_selected} item ({'directory' if selected[0].is_dir else 'file'} '{selected[0].name}')")
        else:
            imgui.text(f"  Selected {num_selected} items")

        return cancelled, closed

    def draw_listing(self, leave_space_for_bottom_bar: bool):
        closed = False
        size = imgui.ImVec2(imgui.get_item_rect_size().x, 0)
        if leave_space_for_bottom_bar:
            button_text_size = imgui.calc_text_size(ifa6.ICON_FA_CIRCLE_XMARK+" Cancel")
            bottom_margin = button_text_size.y+imgui.get_style().frame_padding.y*2+imgui.get_style().item_spacing.y
            size.y = -bottom_margin
        imgui.begin_child("##folder_contents", size=size)
        if self.refreshing and self.new_loc:
            string = 'loading directory...'
            t_size = imgui.calc_text_size(string)
            symbol_size = imgui.calc_text_size("x").y
            spinner_radii = [x/22*symbol_size for x in [22, 16, 10]]
            lw = 3.5/22*symbol_size
            tot_height = t_size.y+2*spinner_radii[0]+lw
            imgui.set_cursor_pos(((imgui.get_content_region_avail().x - t_size.x)/2, (imgui.get_content_region_avail().y - tot_height)/2))
            imgui.text(string)
            imgui.set_cursor_pos_x((imgui.get_content_region_avail().x - 2*spinner_radii[0]+lw)/2)
            imspinner.spinner_ang_triple(f'loadingSpinner', *spinner_radii, lw, c1=imgui.get_style().color_(imgui.Col_.text_selected_bg), c2=imgui.get_style().color_(imgui.Col_.text), c3=imgui.get_style().color_(imgui.Col_.text_selected_bg))
        elif self.msg:
            imgui.text_wrapped(self.msg)
        else:
            # do we have context menus?
            # not for special paths (which are strings) and not for share overviews of a network computer
            net_comps = file_actions.get_net_computer(self.loc) if platform.os==platform.Os.Windows else None
            has_context_menu = not isinstance(self.loc, str) and (not net_comps or len(net_comps)>1)

            table_flags = (
                imgui.TableFlags_.scroll_x |
                imgui.TableFlags_.scroll_y |
                imgui.TableFlags_.hideable |
                imgui.TableFlags_.sortable |
                imgui.TableFlags_.sort_multi |
                imgui.TableFlags_.reorderable |
                imgui.TableFlags_.sizing_fixed_fit |
                imgui.TableFlags_.no_host_extend_y
            )
            if imgui.begin_table(f"##folder_list",columns=5+self.allow_multiple,flags=table_flags):
                frame_height = imgui.get_frame_height()

                # Setup
                checkbox_width = frame_height-2*imgui.get_style().frame_padding.y
                if self.allow_multiple:
                    imgui.table_setup_column("Selector", imgui.TableColumnFlags_.no_hide | imgui.TableColumnFlags_.no_sort | imgui.TableColumnFlags_.no_resize | imgui.TableColumnFlags_.no_reorder | imgui.TableColumnFlags_.no_header_label, init_width_or_weight=checkbox_width)  # 0
                imgui.table_setup_column("Name", imgui.TableColumnFlags_.width_stretch | imgui.TableColumnFlags_.default_sort | imgui.TableColumnFlags_.no_hide)  # 1
                imgui.table_setup_column("Date created", imgui.TableColumnFlags_.default_hide)  # 2
                imgui.table_setup_column("Date modified")  # 3
                imgui.table_setup_column("Type")  # 4
                imgui.table_setup_column("Size")  # 5
                imgui.table_setup_scroll_freeze(int(self.allow_multiple), 1)  # Sticky column headers and selector row

                with self.items_lock:
                    sort_specs = imgui.table_get_sort_specs()
                    self.sort_items(sort_specs)

                    # Headers
                    imgui.table_headers_row()
                    # set up checkbox column: reflects whether all, some or none of visible items are selected, and allows selecting all or none
                    if self.allow_multiple:
                        imgui.table_set_column_index(0)
                        num_selected = sum([self.selected[iid] for iid in self.selected])
                        # determine state
                        if self.predicate_selectable:
                            num_items = sum([self.predicate_selectable(iid) for iid in self.items])
                        else:
                            num_items = len(self.items)
                        if num_selected==0:
                            # none selected
                            multi_selected_state = -1
                        elif num_selected==num_items:
                            # all selected
                            multi_selected_state = 1
                        else:
                            # some selected
                            multi_selected_state = 0

                        if multi_selected_state==0:
                            imgui.internal.push_item_flag(imgui.internal.ItemFlags_.mixed_value, True)
                        clicked, new_state = gui_utils.my_checkbox("##header_checkbox", multi_selected_state==1, frame_size=(0,0), frame_padding_override=(imgui.get_style().frame_padding.x/2,0), do_vertical_align=False)
                        if multi_selected_state==0:
                            imgui.internal.pop_item_flag()

                        if clicked:
                            utils.set_all(self.selected, new_state, subset=self.sorted_items, predicate=self.predicate_selectable)


                    # Loop rows
                    override_color = self.accent_color is not None and self.bg_color is not None
                    if override_color:
                        a=.4
                        style_selected_row = (*tuple(a*x+(1-a)*y for x,y in zip(self.accent_color[:3],self.bg_color[:3])), 1.)
                        a=.2
                        style_hovered_row  = (*tuple(a*x+(1-a)*y for x,y in zip(self.accent_color[:3],self.bg_color[:3])), 1.)
                    any_selectable_clicked = False
                    new_loc = None
                    last_y = None
                    if self.sorted_items and self.last_clicked_id not in self.sorted_items:
                        # default to topmost if last_clicked unknown, or no longer on screen due to filter
                        self.last_clicked_id = self.sorted_items[0]
                    items_to_display = [iid for iid in self.sorted_items if not self.predicate_showable or self.predicate_showable(iid)]
                    clipper = imgui.ListClipper()
                    clipper.begin(len(items_to_display))
                    while clipper.step():
                        for iid in items_to_display[clipper.display_start:clipper.display_end]:
                            imgui.table_next_row()

                            selectable_clicked = False
                            checkbox_clicked, checkbox_hovered, checkbox_out = False, False, False
                            has_drawn_hitbox = False

                            disable_item = self.predicate_selectable and not self.predicate_selectable(iid)
                            if disable_item:
                                imgui.begin_disabled()

                            for ci in range(5+self.allow_multiple):
                                if not (imgui.table_get_column_flags(ci) & imgui.TableColumnFlags_.is_enabled):
                                    continue
                                imgui.table_set_column_index(ci)

                                # Row hitbox
                                if not has_drawn_hitbox:
                                    # hitbox needs to be drawn before anything else on the row so that, together with imgui.set_item_allow_overlap(), hovering button
                                    # or checkbox on the row will still be correctly detected.
                                    # this is super finicky, but works. The below together with using a height of frame_height+cell_padding_y
                                    # makes the table row only cell_padding_y/2 longer. The whole row is highlighted correctly
                                    cell_padding_y = imgui.get_style().cell_padding.y
                                    cur_pos_y = imgui.get_cursor_pos_y()
                                    imgui.set_cursor_pos_y(cur_pos_y - cell_padding_y/2)
                                    imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                                    imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,0.))
                                    imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,cell_padding_y))
                                    if override_color:
                                        # make selectable completely transparent
                                        imgui.push_style_color(imgui.Col_.header_active , (0., 0., 0., 0.))
                                        imgui.push_style_color(imgui.Col_.header        , (0., 0., 0., 0.))
                                        imgui.push_style_color(imgui.Col_.header_hovered, (0., 0., 0., 0.))
                                    selectable_clicked, selectable_out = imgui.selectable(f"##{iid}_hitbox", self.selected[iid], flags=imgui.SelectableFlags_.span_all_columns|imgui.SelectableFlags_.allow_overlap|imgui.internal.SelectableFlagsPrivate_.select_on_click, size=(0,frame_height+cell_padding_y))
                                    # instead override table row background color
                                    if override_color:
                                        if selectable_out:
                                            imgui.table_set_bg_color(imgui.TableBgTarget_.row_bg0, imgui.color_convert_float4_to_u32(style_selected_row))
                                        elif imgui.is_item_hovered():
                                            imgui.table_set_bg_color(imgui.TableBgTarget_.row_bg0, imgui.color_convert_float4_to_u32(style_hovered_row))
                                        imgui.pop_style_color(3)
                                    imgui.pop_style_var(3)
                                    imgui.set_cursor_pos_y(cur_pos_y)   # instead of imgui.same_line(), we just need this part of its effect
                                    selectable_right_clicked, _ = gui_utils.handle_item_hitbox_events(iid, self.selected, context_menu=lambda _: self._item_context_menu([iid for iid in self.sorted_items if self.selected[iid]]) if has_context_menu else None)
                                    has_drawn_hitbox = True

                                if ci==int(self.allow_multiple):
                                    # (Invisible) button because it aligns the following draw calls to center vertically
                                    imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                                    imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,imgui.get_style().frame_padding.y))
                                    imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,imgui.get_style().item_spacing.y))
                                    imgui.push_style_color(imgui.Col_.button, (0.,0.,0.,0.))
                                    imgui.button(f"##{iid}_id", size=(imgui.FLT_MIN,0))
                                    imgui.pop_style_color()
                                    imgui.pop_style_var(3)

                                    imgui.same_line()

                                match ci+int(not self.allow_multiple):
                                    case 0:
                                        # Selector
                                        if disable_item:
                                            checkbox_clicked, checkbox_out, checkbox_hovered = False, False, False
                                        else:
                                            checkbox_clicked, checkbox_out = gui_utils.my_checkbox(f"##{iid}_selected", self.selected[iid], frame_size=(0,0), frame_padding_override=(imgui.get_style().frame_padding.x/2,imgui.get_style().frame_padding.y))
                                            checkbox_hovered = imgui.is_item_hovered()
                                    case 1:
                                        # Name
                                        imgui.set_cursor_pos_x(imgui.get_cursor_pos_x() - imgui.calc_text_size(DIR_ICON).x/3)
                                        imgui.text(self.items[iid].display_name)
                                    case 2:
                                        # Date created
                                        if self.items[iid].ctime_str:
                                            imgui.text(self.items[iid].ctime_str)
                                    case 3:
                                        # Date modified
                                        if self.items[iid].mtime_str:
                                            imgui.text(self.items[iid].mtime_str)
                                    case 4:
                                        # Type
                                        if self.items[iid].mime_type:
                                            disp_str = utils.trim_str(self.items[iid].mime_type,20)
                                            imgui.text(disp_str)
                                            if disp_str!=self.items[iid].mime_type:
                                                gui_utils.draw_hover_text(self.items[iid].mime_type, text='')
                                    case 5:
                                        # Size
                                        if self.items[iid].size_str:
                                            imgui.text(self.items[iid].size_str)

                            if disable_item:
                                imgui.end_disabled()
                            else:
                                last_y = imgui.get_cursor_screen_pos().y

                            # handle selection logic
                            # NB: the part of this logic that has to do with right-clicks is in handle_item_hitbox_events()
                            # NB: any_selectable_clicked is just for handling clicks not on any item
                            any_selectable_clicked = any_selectable_clicked or selectable_clicked or selectable_right_clicked
                            self.last_clicked_id = gui_utils.selectable_item_logic(
                                iid, self.selected, self.last_clicked_id, self.sorted_items,
                                selectable_clicked, selectable_out, allow_multiple=self.allow_multiple,
                                overlayed_hovered=checkbox_hovered, overlayed_clicked=checkbox_clicked, new_overlayed_state=checkbox_out
                                )

                            # further deal with doubleclick on item
                            if selectable_clicked and not checkbox_hovered: # don't enter this branch if interaction is with checkbox on the table row
                                if not imgui.get_io().key_ctrl and not imgui.get_io().key_shift and imgui.is_mouse_double_clicked(imgui.MouseButton_.left):
                                    if self.items[iid].is_dir:
                                        new_loc = self.items[iid].full_path
                                        break
                                    else:
                                        utils.set_all(self.selected, False)
                                        self.selected[iid] = True
                                        imgui.close_current_popup()
                                        closed = True

                    # handle action keys
                    if not self.popup_stack and not self.disable_keyboard_navigation and not imgui.get_io().want_text_input:
                        # no keyboard navigation in this GUI if a popup is open or key input taken elsewhere
                        selected_ids = [iid for iid in self.sorted_items if self.selected[iid]]
                        if len(selected_ids)==1:
                            if imgui.is_key_pressed(imgui.Key.f2, repeat=False):
                                self.dialog_provider.show_rename_path_dialog(self.items[selected_ids[0]].full_path)
                            if self.items[selected_ids[0]].is_dir and imgui.is_key_pressed(imgui.Key.enter, repeat=False):
                                new_loc = self.items[selected_ids[0]].full_path

                if new_loc:
                    self.goto(new_loc)
                if last_y is None:
                    last_y = imgui.get_cursor_screen_pos().y
                imgui.end_table()

                # handle click in table area outside header+contents:
                # deselect all, and if right click, show popup
                # check mouse is below bottom of last drawn row so that clicking on the one pixel empty space between selectables
                # does not cause everything to unselect or popup to open
                if not any_selectable_clicked and imgui.get_io().mouse_pos.y>last_y:
                    if imgui.is_item_clicked(imgui.MouseButton_.left):  # left mouse click (NB: table header is not signalled by is_item_clicked(), so this works correctly)
                        utils.set_all(self.selected, False)
                    # show menu when right-clicking the empty space
                    if imgui.is_item_clicked(imgui.MouseButton_.right): # NB: mouse down
                        utils.set_all(self.selected, False)  # deselect on right mouse click as well
                    if has_context_menu and imgui.begin_popup_context_item("##file_list_context"):   # NB: mouse up
                        if imgui.selectable(f"New folder##button", False)[0]:
                            self.dialog_provider.show_new_folder_dialog(self.loc)
                        imgui.end_popup()

        imgui.end_child()
        return closed

    def _item_context_menu(self, iids: list[int]):
        disabled = len(iids)!=1
        if disabled:
            imgui.begin_disabled()
        if imgui.selectable(f"Rename##button", False)[0]:
            self.dialog_provider.show_rename_path_dialog(self.items[iids[0]].full_path)
        if disabled:
            imgui.end_disabled()
        if imgui.selectable(f"New folder here##button", False)[0]:
            self.dialog_provider.show_new_folder_dialog(self.items[iids[0]].full_path.parent)

    def tick(self):
        # Auto refresh
        if not self.refreshing and (self.elapsed>2 or imgui.is_key_pressed(imgui.Key.f5)):
            self.refresh()

        # Setup popup
        if not imgui.is_popup_open(self.title):
            imgui.open_popup(self.title)
        opened = 1
        size = imgui.get_io().display_size
        size.x *= .7
        size.y *= .7
        imgui.set_next_window_size(size, cond=imgui.Cond_.appearing)
        if imgui.begin_popup_modal(self.title, True, flags=self.default_flags)[0]:
            cancelled = closed = gui_utils.close_weak_popup(check_click_outside=False)
            cancelled2, closed2 = self.draw()
            cancelled = cancelled or cancelled2
            closed    = closed    or closed2
        else:
            opened = 0
            cancelled = closed = True
        if closed:
            if not cancelled and self.callback:
                if self._dir_picker_is_current_dir_selected():
                    selected = [self.loc]
                else:
                    with self.items_lock:
                        selected = [self.items[iid].full_path for iid in self.items if iid in self.selected and self.selected[iid]]
                self.callback(selected if selected else None)

        gui_utils.handle_popup_stack(self.popup_stack)
        return opened, closed

    def sort_items(self, sort_specs_in: imgui.TableSortSpecs):
        if sort_specs_in.specs_dirty or self.require_sort:
            ids = list(self.items)
            sort_specs = [sort_specs_in.get_specs(i) for i in range(sort_specs_in.specs_count)]
            for sort_spec in reversed(sort_specs):
                match sort_spec.column_index+int(not self.allow_multiple):
                    case 2:     # Date created
                        key = lambda iid: self.items[iid].ctime
                    case 3:     # Date modified
                        key = lambda iid: self.items[iid].mtime
                    case 4:     # Type
                        key = lambda iid: m if (m:=self.items[iid].mime_type) else ''
                    case 5:     # Size
                        key = lambda iid: self.items[iid].size

                    case _:     # Name and all others
                        key = natsort.os_sort_keygen(key=lambda iid: self.items[iid].full_path)

                ids.sort(key=key, reverse=sort_spec.get_sort_direction()==imgui.SortDirection.descending)

            # finally, always sort dirs first
            ids.sort(key=lambda iid: self.items[iid].is_dir, reverse=True)
            self.sorted_items = ids

            # apply filter, if any
            if self.filter_box_text:
                search = self.filter_box_text.casefold()
                def key(iid):
                    item = self.items[iid]
                    return search in item.display_name.casefold()
                self.sorted_items = list(filter(key, self.sorted_items))

            # we're done
            sort_specs_in.specs_dirty = False
            self.require_sort = False

class DirPicker(FilePicker):
    def __init__(self, title="Directory picker", start_dir: str | pathlib.Path = None, callback: typing.Callable = None, allow_multiple = True, custom_popup_flags=0):
        super().__init__(title=title, dir_picker=True, start_dir=start_dir, callback=callback, allow_multiple=allow_multiple, custom_popup_flags=custom_popup_flags)

class DialogProvider:
    def __init__(self, gui, action_provider):
        self.gui = gui
        self.action_provider = action_provider
    def show_new_folder_dialog(self, parent=pathlib.Path):
        new_folder_name = 'New folder'
        width = imgui.calc_text_size('x').x*35
        setup_done = False
        def _new_folder_popup():
            nonlocal new_folder_name, setup_done
            if imgui.begin_table("##new_folder",2):
                imgui.table_setup_column("##new_folder_left", imgui.TableColumnFlags_.width_fixed)
                imgui.table_setup_column("##new_folder_right", imgui.TableColumnFlags_.width_stretch)
                imgui.table_next_row()
                imgui.table_next_column()
                imgui.align_text_to_frame_padding()
                imgui.text("Folder name")
                imgui.table_next_column()
                imgui.set_next_item_width(width)
                if not setup_done:
                    imgui.set_keyboard_focus_here()
                    setup_done = True
                _,new_folder_name = imgui.input_text("##new_folder_name", new_folder_name)
                imgui.end_table()
            return 0 if imgui.is_key_released(imgui.Key.enter) else None
        buttons = {
            ifa6.ICON_FA_CHECK+" Make folder": lambda: self.action_provider('make_dir',parent/new_folder_name),
            ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
        }
        gui_utils.push_popup(self.gui, lambda: gui_utils.popup("Make folder", _new_folder_popup, buttons = buttons, closable=True))

    def show_rename_path_dialog(self, item: pathlib.Path):
        item_name = item.name
        width = imgui.calc_text_size('x').x*(len(item_name)+15)
        setup_done = False
        def _rename_item_popup():
            nonlocal item_name, setup_done
            if imgui.begin_table("##rename_item",2):
                imgui.table_setup_column("##rename_item_left", imgui.TableColumnFlags_.width_fixed)
                imgui.table_setup_column("##rename_item_right", imgui.TableColumnFlags_.width_stretch)
                imgui.table_next_row()
                imgui.table_next_column()
                imgui.align_text_to_frame_padding()
                imgui.text("Item name")
                imgui.table_next_column()
                imgui.set_next_item_width(width)
                if not setup_done:
                    imgui.set_keyboard_focus_here()
                    setup_done = True
                _,item_name = imgui.input_text("##new_rename_item", item_name)
                imgui.end_table()
            return 0 if imgui.is_key_released(imgui.Key.enter) else None
        buttons = {
            ifa6.ICON_FA_CHECK+" Rename": lambda: self.action_provider('rename_path', item, item.parent / item_name),
            ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
        }
        gui_utils.push_popup(self.gui, lambda: gui_utils.popup("Rename item", _rename_item_popup, buttons = buttons, closable=True))


import enum
import json
import pathlib
import dataclasses
import typing
import types
import numpy as np
import os
import colorsys


def hex_to_rgba_0_1(hex):
    r = int(hex[1:3], base=16) / 255
    g = int(hex[3:5], base=16) / 255
    b = int(hex[5:7], base=16) / 255
    if len(hex) > 7:
        a = int(hex[7:9], base=16) / 255
    else:
        a = 1.0
    return (r, g, b, a)

def rgba_0_1_to_hex(rgba):
    r = "%.2x" % int(rgba[0] * 255)
    g = "%.2x" % int(rgba[1] * 255)
    b = "%.2x" % int(rgba[2] * 255)
    if len(rgba) > 3:
        a = "%.2x" % int(rgba[3] * 255)
    else:
        a = "FF"
    return f"#{r}{g}{b}{a}"

def get_colors(n_colors: int, saturation: float, value: float) -> list[tuple[float, float, float]]:
    color_steps = 1/(n_colors+1)
    return [colorsys.hsv_to_rgb(i*color_steps, saturation, value) for i in range(n_colors)]

def get_hour_minutes_seconds_ms(dur_seconds: float) -> tuple[float, float, float, float]:
    hours, remainder = divmod(dur_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    seconds, ms      = divmod(seconds, 1)
    return hours, minutes, seconds, ms
def format_duration(dur: float, show_ms: bool) -> str:
    hours, minutes, seconds, ms = get_hour_minutes_seconds_ms(dur)
    if round(ms,3)==1.:
        # prevent getting timecode x:xx:xx.1000
        hours, minutes, seconds, ms = get_hour_minutes_seconds_ms(round(dur))
    dur_str = f'{int(hours)}:{int(minutes):02d}:{int(seconds):02d}'
    if show_ms:
        dur_str += f'.{ms*1000:03.0f}'
    return dur_str


class AutoName(enum.Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name.strip("_").replace("__", "-").replace("_", " ")


@dataclasses.dataclass
class CustomTypeEntry:
    type        : typing.Type
    reg_name    : str
    to_json     : typing.Callable
    from_json   : typing.Callable

CUSTOM_TYPE_REGISTRY = []
def register_type(entry: CustomTypeEntry):
    CUSTOM_TYPE_REGISTRY.append(entry)
register_type(CustomTypeEntry(pathlib.Path,'pathlib.Path',str,lambda x: pathlib.Path(x)))
register_type(CustomTypeEntry(set,'builtin.set',list,lambda x: set(x)))

class CustomTypeEncoder(json.JSONEncoder):
    def default(self, obj):
        for t in CUSTOM_TYPE_REGISTRY:
            if isinstance(obj, t.type):
                return {t.reg_name: t.to_json(obj)}
        return json.JSONEncoder.default(self, obj)

def json_reconstitute(d):
    for t in CUSTOM_TYPE_REGISTRY:
        if t.reg_name in d:
            return t.from_json(d[t.reg_name])
    return d


def cartesian_product(*arrays):
    ndim = len(arrays)
    return (np.stack(np.meshgrid(*arrays), axis=-1).reshape(-1, ndim))


def fast_scandir(dirname) -> list[pathlib.Path]:
    if not dirname.is_dir():
        return []
    subfolders= [pathlib.Path(f.path) for f in os.scandir(dirname) if f.is_dir()]
    for dirname in list(subfolders):
        subfolders.extend(fast_scandir(dirname))
    return subfolders

def unpack_none_union(annotation: typing.Type) -> tuple[typing.Type, bool]:
    # below handles both types.Optional and direct unions with None
    if typing.get_origin(annotation) in [typing.Union, types.UnionType] and (args:=typing.get_args(annotation))[-1]==types.NoneType:
        return typing.Union[args[:-1]], True
    else:
        return annotation, False


def set_all(inp: dict[int, bool], value, subset: list[int] = None, predicate: typing.Callable[[int], bool] = None):
    if subset is None:
        subset = (r for r in inp)
    for r in subset:
        if r in inp and (not predicate or predicate(r)):
            inp[r] = value


def trim_str(text: str, length=None, till_newline=True, newline_ellipsis=False):
    if text and till_newline:
        temp = text.splitlines()
        if temp:
            text = temp[0]
        if len(temp)>1 and newline_ellipsis:
            text += '..'
    if length:
        text = (text[:length-2] + '..') if len(text) > length else text
    return text


import enum

from . import utils

class EyeTracker(utils.AutoName):
    AdHawk_MindLink = enum.auto()
    Pupil_Core      = enum.auto()
    Pupil_Invisible = enum.auto()
    Pupil_Neon      = enum.auto()
    SeeTrue_STONE   = enum.auto()
    SMI_ETG         = enum.auto()
    Tobii_Glasses_2 = enum.auto()
    Tobii_Glasses_3 = enum.auto()
    Unknown         = enum.auto()
eye_tracker_names = [x.value for x in EyeTracker if x!=EyeTracker.Unknown]

EyeTracker.AdHawk_MindLink.color = utils.hex_to_rgba_0_1("#001D7A")
EyeTracker.Pupil_Core     .color = utils.hex_to_rgba_0_1("#E6194B")
EyeTracker.Pupil_Invisible.color = utils.hex_to_rgba_0_1("#3CB44B")
EyeTracker.Pupil_Neon     .color = utils.hex_to_rgba_0_1("#C6B41E")
EyeTracker.SeeTrue_STONE  .color = utils.hex_to_rgba_0_1("#911EB4")
EyeTracker.SMI_ETG        .color = utils.hex_to_rgba_0_1("#4363D8")
EyeTracker.Tobii_Glasses_2.color = utils.hex_to_rgba_0_1("#F58231")
EyeTracker.Tobii_Glasses_3.color = utils.hex_to_rgba_0_1("#F032E6")
EyeTracker.Unknown        .color = utils.hex_to_rgba_0_1("#393939")

def string_to_enum(device: str) -> EyeTracker:
    if isinstance(device, EyeTracker):
        return device

    if isinstance(device, str):
        if hasattr(EyeTracker, device):
            return getattr(EyeTracker, device)
        elif device in [e.value for e in EyeTracker]:
            return EyeTracker(device)
        else:
            raise ValueError(f"The string '{device}' is not a known eye tracker type. Known types: {[e.value for e in EyeTracker]}")
    else:
        raise ValueError(f"The variable 'device' should be a string with one of the following values: {[e.value for e in EyeTracker]}")


utils.register_type(utils.CustomTypeEntry(EyeTracker,'__enum.EyeTracker__',str, lambda x: getattr(EyeTracker, x.split('.')[1])))


import dataclasses
import pathlib
import typing
import shutil
import json
import os

from . import utils, video_utils


@dataclasses.dataclass
class Recording:
    default_json_file_name      : typing.ClassVar[str] = 'recording_info.json'

    name                        : str
    video_file                  : str
    source_directory            : pathlib.Path
    working_directory           : pathlib.Path  = ""
    duration                    : int           = None

    def store_as_json(self, path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path /= self.default_json_file_name
        with open(path, 'w') as f:
            # remove any crap potentially added by subclasses
            to_dump = dataclasses.asdict(self)
            to_dump = {k:to_dump[k] for k in to_dump if k in Recording.__annotations__ and k not in ['working_directory']}      # working_directory will be loaded as the provided path, and shouldn't be stored
            # dump to file
            json.dump(to_dump, f, cls=utils.CustomTypeEncoder, indent=2)

    @staticmethod
    def load_from_json(path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path /= Recording.default_json_file_name
        with open(path, 'r') as f:
            return Recording(**json.load(f, object_hook=utils.json_reconstitute), working_directory=path.parent)


    def get_video_path(self):
        vid = self.working_directory / self.video_file
        if not vid.is_file():
            if not self.source_directory.is_absolute():
                vid = (self.working_directory / self.source_directory / self.video_file).resolve()
            else:
                vid = self.source_directory / self.video_file
        return vid



def do_import(rec_info: Recording, cam_cal_file: str|pathlib.Path=None, copy_video=True, source_dir_as_relative_path = False):
    if not rec_info.working_directory:
        raise ValueError('working_directory must be set on the rec_info object')
    rec_info.working_directory = pathlib.Path(rec_info.working_directory)
    ifile = (rec_info.source_directory / rec_info.video_file)
    if not ifile.is_file():
        raise FileNotFoundError(f'The camera recording file {ifile} was not found')
    print(f'processing: {rec_info.video_file} -> {rec_info.working_directory}')

    if not rec_info.working_directory.is_dir():
        rec_info.working_directory.mkdir()

    if copy_video:
        ofile = rec_info.working_directory / rec_info.video_file
        print('  Copy video file...')
        shutil.copy2(ifile, ofile)

    # also get its calibration
    print('  Getting camera calibration...')
    if cam_cal_file is not None:
        shutil.copy2(str(cam_cal_file), str(rec_info.working_directory / 'calibration.xml'))
    else:
        print('  !! No camera calibration provided! Defaulting to hardcoded')

    # and frame timestamps
    print('  Getting frame timestamps...')
    ts = video_utils.get_frame_timestamps_from_video(rec_info.get_video_path())
    ts.to_csv(str(rec_info.working_directory / 'frameTimestamps.tsv'), sep='\t')
    rec_info.duration = ts.timestamp.iat[-1]-ts.timestamp.iat[0]

    # store recording info to folder
    if source_dir_as_relative_path:
        rec_info.source_directory = pathlib.Path(os.path.relpath(rec_info.source_directory,rec_info.working_directory))
    rec_info.store_as_json(rec_info.working_directory)

    return rec_info


import dataclasses
import datetime
import typing
import threading
from imgui_bundle import imgui, icons_fontawesome_6 as ifa6

from . import utils as gui_utils
from .. import camera_recording, eyetracker, recording, utils



@dataclasses.dataclass
class Filter:
    fun: typing.Callable[[int|str, recording.Recording], bool]
    invert = False
    id: int = dataclasses.field(init=False)

    def __post_init__(self):
        self.id = id(self)

class ColumnSpec(typing.NamedTuple):
    position: int
    name: str
    flags: int
    display_func:  typing.Callable[[recording.Recording|camera_recording.Recording],None]
    sort_key_func: typing.Callable[[int],typing.Any]
    header_lbl: str|None=None   # if set, different string than name is used for the column header. Works only for non-angled headers

class RecordingTable:
    def __init__(self,
                 recordings: dict[int|str, recording.Recording|camera_recording.Recording],
            recordings_lock: threading.Lock,
        selected_recordings: dict[int|str, bool]|None,
        extra_columns: list[ColumnSpec] = None,
        get_rec_fun: typing.Callable[[typing.Any], recording.Recording|camera_recording.Recording] = None,
        item_context_callback: typing.Callable[[int], bool] = None,
        empty_context_callback: typing.Callable[[],None] = None,
        item_remove_callback: typing.Callable[[int], None] = None
        ):

        self.recordings = recordings
        self.recordings_lock = recordings_lock
        self.selected_recordings = selected_recordings
        self.has_selected_recordings = self.selected_recordings is not None
        if not self.has_selected_recordings:
            # make an internal one
            self.selected_recordings = {iid:False for iid in self.recordings}

        self.item_context_callback  = item_context_callback
        self.empty_context_callback = empty_context_callback
        self.item_remove_callback   = item_remove_callback
        self.get_rec_fun            = lambda rec: rec
        if get_rec_fun is not None:
            self.get_rec_fun        = get_rec_fun
        self.dont_show_empty        = False

        self.is_drag_drop_source    = False
        self.drag_drop_id           = None

        self.sorted_recordings_ids: list[int] = []
        self.last_clicked_id: int = None
        self.require_sort: bool = True
        self.filters: list[Filter] = []
        self.filter_box_text: str = ""

        self._last_y = None
        self._has_scroll_x = None

        self._columns: list[ColumnSpec] = []
        self._show_hide_commands: dict[int,bool] = {}
        self.build_columns(extra_columns)

        self._eye_tracker_label_width: float = None
        self.table_flags: int = (
            imgui.TableFlags_.scroll_x |
            imgui.TableFlags_.scroll_y |
            imgui.TableFlags_.hideable |
            imgui.TableFlags_.sortable |
            imgui.TableFlags_.resizable |
            imgui.TableFlags_.sort_multi |
            imgui.TableFlags_.reorderable |
            imgui.TableFlags_.row_bg |
            imgui.TableFlags_.sizing_fixed_fit |
            imgui.TableFlags_.no_host_extend_y |
            imgui.TableFlags_.no_borders_in_body_until_resize
        )

    def build_columns(self, extra_columns: list[ColumnSpec] = None):
        col_names = ifa6.ICON_FA_EYE+" Eye Tracker", ifa6.ICON_FA_SIGNATURE+" Name", ifa6.ICON_FA_USER_TIE+" Participant", ifa6.ICON_FA_CLIPBOARD+" Project", ifa6.ICON_FA_STOPWATCH+" Duration", ifa6.ICON_FA_CLOCK+" Recording Start", ifa6.ICON_FA_FOLDER+" Working Directory", ifa6.ICON_FA_FOLDER+" Source Directory", ifa6.ICON_FA_TAGS+" Firmware Version", ifa6.ICON_FA_BARCODE+" Glasses Serial", ifa6.ICON_FA_BARCODE+" Recording Unit Serial", ifa6.ICON_FA_TAGS+" Recording Software Version", ifa6.ICON_FA_BARCODE+" Scene Camera Serial", ifa6.ICON_FA_CAMERA+" Video File"
        if self.has_selected_recordings:
            col_names = (ifa6.ICON_FA_SQUARE_CHECK+" Selector",)+col_names
        i_def_col = 0
        i_col = 0
        extra_columns_pos = [x.position for x in extra_columns] if extra_columns else []
        has_user_default_sort = False if extra_columns is None else any(((e.flags & imgui.TableColumnFlags_.default_sort) for e in extra_columns))
        self._columns = []
        while True:
            if i_col in extra_columns_pos:
                self._columns.append(extra_columns[extra_columns_pos.index(i_col)])
            else:
                if i_def_col>=len(col_names):
                    break
                col = self._get_column(col_names[i_def_col], i_col)
                if has_user_default_sort:
                    # ensure default sort is not set
                    col = col._replace(flags=col.flags & ~imgui.TableColumnFlags_.default_sort)
                self._columns.append(col)
                i_def_col += 1
            i_col += 1
        self._show_hide_commands.clear()

    def show_hide_columns(self, show_hide: dict[str, bool]):
        for c in show_hide:
            matching = [c in col.name for col in self._columns]
            n_matching = sum(matching)
            if n_matching==0:
                raise ValueError(f'No column found whose name contains "{c}"')
            elif n_matching>1:
                raise ValueError(f'More than one column found whose name contains "{c}": {[col.name for i,col in enumerate(self._columns) if matching[i]]}')
            c_idx = matching.index(True)
            self._show_hide_commands[c_idx] = show_hide[c]

    def _get_column(self, name: str, position: int):
        flags = imgui.TableColumnFlags_.default_hide | imgui.TableColumnFlags_.no_resize    # most columns use this one
        match name[2:]:
            case "Selector":
                flags = imgui.TableColumnFlags_.no_hide | imgui.TableColumnFlags_.no_sort | imgui.TableColumnFlags_.no_resize | imgui.TableColumnFlags_.no_reorder
                display_func = None
                sort_key_func= None
            case "Eye Tracker":
                flags = imgui.TableColumnFlags_.no_resize
                display_func = lambda rec: self.draw_eye_tracker_widget(self.get_rec_fun(rec), align=True)
                sort_key_func= lambda iid: (r.eye_tracker.value if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'Camera').lower()
            case "Name":
                flags = imgui.TableColumnFlags_.default_sort | imgui.TableColumnFlags_.no_hide | imgui.TableColumnFlags_.no_resize
                display_func = None # special case
                sort_key_func= lambda iid: self.get_rec_fun(self.recordings[iid]).name.lower()
            case "Participant":
                flags = imgui.TableColumnFlags_.no_resize
                display_func = lambda rec: imgui.text(r.participant or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.participant if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Project":
                display_func = lambda rec: imgui.text(r.project or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.project if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Duration":
                flags = imgui.TableColumnFlags_.no_resize
                display_func = lambda rec: imgui.text("Unknown" if (d:=self.get_rec_fun(rec).duration) is None else str(datetime.timedelta(seconds=d//1000)))
                sort_key_func= lambda iid: 0 if (d:=self.get_rec_fun(self.recordings[iid]).duration) is None else d
            case "Recording Start":
                display_func = lambda rec: imgui.text(r.start_time.display or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: r.start_time.value if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 0
            case "Working Directory":
                display_func = lambda rec: self.draw_working_directory(self.get_rec_fun(rec))
                sort_key_func= lambda iid: self.get_rec_fun(self.recordings[iid]).working_directory.name.lower()
            case "Source Directory":
                display_func = lambda rec: self.draw_source_directory(self.get_rec_fun(rec))
                sort_key_func= lambda iid: str(self.get_rec_fun(self.recordings[iid]).source_directory).lower()
            case "Firmware Version":
                display_func = lambda rec: imgui.text(r.firmware_version or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.firmware_version if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Glasses Serial":
                display_func = lambda rec: imgui.text(r.glasses_serial or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.glasses_serial if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Recording Unit Serial":
                display_func = lambda rec: imgui.text(r.recording_unit_serial or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.recording_unit_serial if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Recording Software Version":
                display_func = lambda rec: imgui.text(r.recording_software_version or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.recording_software_version if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Scene Camera Serial":
                display_func = lambda rec: imgui.text(r.scene_camera_serial or "Unknown") if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else None
                sort_key_func= lambda iid: (r.scene_camera_serial if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else 'zzzzz').lower()
            case "Video File":
                display_func = lambda rec: imgui.text((r.scene_video_file if isinstance(r:=self.get_rec_fun(rec), recording.Recording) else r.video_file) or "Unknown")
                sort_key_func= lambda iid: (r.scene_video_file if isinstance(r:=self.get_rec_fun(self.recordings[iid]), recording.Recording) else r.video_file).lower()
            case _:
                raise NotImplementedError()

        return ColumnSpec(position, name, flags, display_func, sort_key_func, name[2:])

    def set_act_as_drag_drop_source(self, enabled: bool):
        self.is_drag_drop_source = enabled

    def add_filter(self, filt: Filter):
        self.filters.append(filt)
        self.require_sort = True

    def remove_filter(self, iid):
        for i, search in enumerate(self.filters):
            if search.id == iid:
                self.filters.pop(i)
        self.require_sort = True

    def font_changed(self):
        self._eye_tracker_label_width = None

    def set_local_item_remover(self):
        self.item_remove_callback = self.remove_recording

    def draw(self, accent_color: tuple[float] = None, bg_color: tuple[float] = None, style_color_recording_name: bool = False, limit_outer_size: bool = False):
        if self.dont_show_empty:
            with self.recordings_lock:
                if not self.recordings:
                    imgui.text_wrapped('There are no recordings')
                    return
        outer_size = imgui.ImVec2(0,0)
        if limit_outer_size and self._last_y is not None:
            outer_size.y = self._last_y+imgui.get_style().item_spacing.y
            if self._has_scroll_x:
                outer_size.y += imgui.get_style().scrollbar_size
        if imgui.begin_table(
            f"##recording_list",
            columns=len(self._columns),
            flags=self.table_flags,
            outer_size=outer_size
        ):
            frame_height = imgui.get_frame_height()

            # Setup
            checkbox_width = frame_height
            has_angled_headers = False
            for c_idx in range(len(self._columns)):
                col = self._columns[c_idx]
                if c_idx==0 and self.has_selected_recordings:
                    imgui.table_setup_column(col.name, col.flags, init_width_or_weight=checkbox_width)
                else:
                    imgui.table_setup_column(col.name, col.flags)
                has_angled_headers = has_angled_headers or (col.flags & imgui.TableColumnFlags_.angled_header)

            # show/hide columns through API
            for c_idx in self._show_hide_commands:
                imgui.table_set_column_enabled(c_idx, self._show_hide_commands[c_idx])
            self._show_hide_commands.clear()

            # Sticky column headers and selector row
            n_row_freeze = 2 if has_angled_headers else 1
            n_col_freeze = 1 if self.has_selected_recordings else 0
            imgui.table_setup_scroll_freeze(n_col_freeze, n_row_freeze)

            with self.recordings_lock:
                if (rs:=set(self.recordings.keys())) != (rss:= set(self.selected_recordings.keys())) or rs!=set(self.sorted_recordings_ids):
                    self.require_sort = True
                    if (new_recs := rs-rss):
                        self.selected_recordings |= {iid:False for iid in new_recs}

                # Sorting
                sort_specs = imgui.table_get_sort_specs()
                sorted_recordings_ids_len = len(self.sorted_recordings_ids)
                self.sort_and_filter_recordings(sort_specs)
                if len(self.sorted_recordings_ids) < sorted_recordings_ids_len:
                    # we've just filtered out some recordings from view. Deselect those
                    # NB: will also be triggered when removing an item, doesn't matter
                    for iid in self.recordings:
                        if iid not in self.sorted_recordings_ids:
                            self.selected_recordings[iid] = False

                # Headers
                if has_angled_headers:
                    imgui.table_angled_headers_row()
                imgui.table_next_row(imgui.TableRowFlags_.headers)
                for c_idx in range(len(self._columns)):
                    if not imgui.table_set_column_index(c_idx):
                        continue
                    if c_idx==0 and self.has_selected_recordings:  # checkbox column: reflects whether all, some or none of visible recordings are selected, and allows selecting all or none
                        # get state
                        num_selected = sum([self.selected_recordings[iid] for iid in self.sorted_recordings_ids])
                        if num_selected==0:
                            # none selected
                            multi_selected_state = -1
                        elif num_selected==len(self.sorted_recordings_ids):
                            # all selected
                            multi_selected_state = 1
                        else:
                            # some selected
                            multi_selected_state = 0

                        if multi_selected_state==0:
                            imgui.internal.push_item_flag(imgui.internal.ItemFlags_.mixed_value, True)
                        clicked, new_state = gui_utils.my_checkbox("##header_checkbox", multi_selected_state==1, frame_size=(0,0), do_vertical_align=False)
                        if multi_selected_state==0:
                            imgui.internal.pop_item_flag()

                        if clicked:
                            utils.set_all(self.selected_recordings, new_state, subset = self.sorted_recordings_ids)
                    else:
                        column_name = self._columns[c_idx].header_lbl if self._columns[c_idx].header_lbl is not None else self._columns[c_idx].name
                        if imgui.table_get_column_flags(c_idx) & imgui.TableColumnFlags_.no_header_label:
                            column_name = '##'+column_name
                        imgui.table_header(column_name)

                # Loop rows
                override_color = accent_color is not None and bg_color is not None
                if override_color:
                    a=.4
                    style_selected_row = (*tuple(a*x+(1-a)*y for x,y in zip(accent_color[:3],bg_color[:3])), 1.)
                    a=.2
                    style_hovered_row  = (*tuple(a*x+(1-a)*y for x,y in zip(accent_color[:3],bg_color[:3])), 1.)
                any_selectable_clicked = False
                if self.sorted_recordings_ids and self.last_clicked_id not in self.sorted_recordings_ids:
                    # default to topmost if last_clicked unknown, or no longer on screen due to filter
                    self.last_clicked_id = self.sorted_recordings_ids[0]
                submitted_drag_drop = False
                for iid in self.sorted_recordings_ids:
                    imgui.table_next_row()

                    num_columns_drawn = 0
                    selectable_clicked = False
                    checkbox_clicked, checkbox_hovered, checkbox_out = False, False, False
                    remove_button_hovered = False
                    has_drawn_hitbox = False
                    should_remove_item = False
                    for c_idx in range(len(self._columns)):
                        if not (imgui.table_get_column_flags(c_idx) & imgui.TableColumnFlags_.is_enabled):
                            continue
                        imgui.table_set_column_index(c_idx)

                        # Row hitbox
                        if not has_drawn_hitbox:
                            # hitbox needs to be drawn before anything else on the row so that, together with imgui.set_item_allow_overlap(), hovering button
                            # or checkbox on the row will still be correctly detected.
                            # this is super finicky, but works. The below together with using a height of frame_height+cell_padding_y
                            # makes the table row only cell_padding_y/2 longer. The whole row is highlighted correctly
                            cell_padding_y = imgui.get_style().cell_padding.y
                            cur_pos_y = imgui.get_cursor_pos_y()
                            imgui.set_cursor_pos_y(cur_pos_y - cell_padding_y/2)
                            imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                            imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,0.))
                            imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,cell_padding_y))
                            if override_color:
                                # make selectable completely transparent
                                imgui.push_style_color(imgui.Col_.header_active , (0., 0., 0., 0.))
                                imgui.push_style_color(imgui.Col_.header        , (0., 0., 0., 0.))
                                imgui.push_style_color(imgui.Col_.header_hovered, (0., 0., 0., 0.))
                            selectable_clicked, selectable_out = imgui.selectable(f"##{iid}_hitbox", self.selected_recordings[iid], flags=imgui.SelectableFlags_.span_all_columns|imgui.SelectableFlags_.allow_overlap|imgui.internal.SelectableFlagsPrivate_.select_on_click, size=(0,frame_height+cell_padding_y))
                            # instead override table row background color, if wanted
                            if override_color:
                                if selectable_out:
                                    imgui.table_set_bg_color(imgui.TableBgTarget_.row_bg0, imgui.color_convert_float4_to_u32(style_selected_row))
                                elif imgui.is_item_hovered():
                                    imgui.table_set_bg_color(imgui.TableBgTarget_.row_bg0, imgui.color_convert_float4_to_u32(style_hovered_row))
                                imgui.pop_style_color(3)
                            imgui.pop_style_var(3)
                            # act as drag/drop source, if wanted
                            if self.is_drag_drop_source and imgui.begin_drag_drop_source(imgui.DragDropFlags_.payload_auto_expire):
                                # Set payload to carry the index of our item (in python, the payload is an int)
                                self.drag_drop_id = dd_id = iid
                                if not isinstance(iid,int):
                                    dd_id = -1

                                imgui.set_drag_drop_payload_py_id("RECORDING", dd_id)
                                submitted_drag_drop = True
                                # Display preview
                                if isinstance(rec:=self.get_rec_fun(self.recordings[iid]), recording.Recording):
                                    imgui.text(rec.name)
                                else:
                                    imgui.text(str(rec.source_directory / rec.video_file))
                                imgui.end_drag_drop_source()

                            imgui.set_cursor_pos_y(cur_pos_y)   # instead of imgui.same_line(), we just need this part of its effect
                            selectable_right_clicked, selectables_edited = gui_utils.handle_item_hitbox_events(iid, self.selected_recordings, self.item_context_callback)
                            self.require_sort |= selectables_edited
                            has_drawn_hitbox = True

                        if num_columns_drawn==1:
                            # (Invisible) button because it aligns the following draw calls to center vertically
                            imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0.)
                            imgui.push_style_var(imgui.StyleVar_.frame_padding    , (0.,imgui.get_style().frame_padding.y))
                            imgui.push_style_var(imgui.StyleVar_.item_spacing     , (0.,imgui.get_style().item_spacing.y))
                            imgui.push_style_color(imgui.Col_.button, (0.,0.,0.,0.))
                            imgui.button(f"##{iid}_id", size=(imgui.FLT_MIN, 0))
                            imgui.pop_style_color()
                            imgui.pop_style_var(3)

                            imgui.same_line()

                        if c_idx==0 and self.has_selected_recordings:
                            # Selector
                            checkbox_clicked, checkbox_out = gui_utils.my_checkbox(f"##{iid}_selected", self.selected_recordings[iid], frame_size=(0,0))
                            checkbox_hovered = imgui.is_item_hovered()
                        elif self._columns[c_idx].header_lbl=="Name":
                            if self.item_remove_callback:
                                if imgui.button(ifa6.ICON_FA_TRASH_CAN+f"##{iid}_remove"):
                                    should_remove_item = True
                                remove_button_hovered = imgui.is_item_hovered()
                                imgui.same_line()
                            self.draw_recording_name_text(self.get_rec_fun(self.recordings[iid]), accent_color if style_color_recording_name else None)
                        else:
                            self._columns[c_idx].display_func(self.recordings[iid])
                        num_columns_drawn+=1

                    # handle item removal
                    if should_remove_item:
                        self.item_remove_callback(iid)
                        self.require_sort = True

                    # handle selection logic
                    # NB: the part of this logic that has to do with right-clicks is in handle_recording_hitbox_events()
                    # NB: any_selectable_clicked is just for handling clicks not on any recording
                    any_selectable_clicked = any_selectable_clicked or selectable_clicked or selectable_right_clicked

                    self.last_clicked_id = gui_utils.selectable_item_logic(
                        iid, self.selected_recordings, self.last_clicked_id, self.sorted_recordings_ids,
                        selectable_clicked, selectable_out, overlayed_hovered=checkbox_hovered or remove_button_hovered,
                        overlayed_clicked=checkbox_clicked, new_overlayed_state=checkbox_out
                        )

                self._last_y = imgui.get_cursor_pos().y
                self._has_scroll_x = imgui.get_current_context().current_window.scrollbar_x
                last_cursor_y = imgui.get_cursor_screen_pos().y
                imgui.end_table()

                if not submitted_drag_drop:
                    self.drag_drop_id = None

                # handle click in table area outside header+contents:
                # deselect all, and if right click, show popup
                # check mouse is below bottom of last drawn row so that clicking on the one pixel empty space between selectables
                # does not cause everything to unselect or popup to open
                if imgui.is_item_clicked(imgui.MouseButton_.left) and not any_selectable_clicked and imgui.get_io().mouse_pos.y>last_cursor_y:  # NB: table header is not signalled by is_item_clicked(), so this works correctly
                    utils.set_all(self.selected_recordings, False)

                # show menu when right-clicking the empty space
                if self.empty_context_callback and imgui.get_io().mouse_pos.y>last_cursor_y and imgui.begin_popup_context_item("##recording_list_context",popup_flags=imgui.PopupFlags_.mouse_button_right | imgui.PopupFlags_.no_open_over_existing_popup):
                    utils.set_all(self.selected_recordings, False)  # deselect on right mouse click as well
                    self.empty_context_callback()
                    imgui.end_popup()

    def remove_recording(self, iid: int|str):
        self.recordings.pop(iid,None)
        self.selected_recordings.pop(iid,None)

    def draw_eye_tracker_widget(self, rec: recording.Recording|camera_recording.Recording, align=False):
        imgui.push_style_var(imgui.StyleVar_.frame_border_size, 0)
        x_padding = 4
        if self._eye_tracker_label_width is None:
            self._eye_tracker_label_width = 0
            for et in list(eyetracker.EyeTracker):
                self._eye_tracker_label_width = max(self._eye_tracker_label_width, imgui.calc_text_size(et.value).x)
            self._eye_tracker_label_width += 2 * x_padding
        if align:
            imgui.begin_group()
            imgui.set_cursor_pos_y(imgui.get_cursor_pos_y() + imgui.get_style().frame_padding.y)

        # prep for drawing widget: determine its size and position and see if visible
        if isinstance(rec, recording.Recording):
            et          = rec.eye_tracker.value
            clr         = rec.eye_tracker.color
        else:
            # camera recording
            et          = 'Camera'
            clr         = (.7, .7, .7, 1.)
        iid         = imgui.get_id(et)
        label_size  = imgui.calc_text_size(et)
        size        = imgui.ImVec2(self._eye_tracker_label_width, label_size.y)
        pos         = imgui.get_cursor_screen_pos()
        bb          = imgui.internal.ImRect(pos, (pos.x+size.x, pos.y+size.y))
        imgui.internal.item_size(size, 0)
        # if visible
        if imgui.internal.item_add(bb, iid):
            # draw frame
            imgui.internal.render_frame(bb.min, bb.max, imgui.color_convert_float4_to_u32(clr), True, imgui.get_style().frame_rounding)
            # draw text on top
            imgui.push_style_color(imgui.Col_.text, (1., 1., 1., 1.))
            imgui.internal.render_text_clipped((bb.min.x+x_padding, bb.min.y), (bb.max.x-x_padding, bb.max.y), et, None, label_size, imgui.get_style().button_text_align, bb)
            imgui.pop_style_color()

        if align:
            imgui.end_group()
        imgui.pop_style_var()

    def draw_recording_name_text(self, rec: recording.Recording|camera_recording.Recording, accent_color: tuple[float] = None):
        if accent_color is not None:
            imgui.text_colored(accent_color, rec.name)
        else:
            imgui.text(rec.name)

    def draw_working_directory(self, rec: recording.Recording|camera_recording.Recording):
        imgui.text(rec.working_directory.name if rec.working_directory else "Unknown")
        if imgui.is_item_hovered():
            if rec.working_directory and rec.working_directory.is_dir():
                text = str(rec.working_directory)
            else:
                text = 'Working directory not created yet'
            gui_utils.draw_tooltip(text)

    def draw_source_directory(self, rec: recording.Recording|camera_recording.Recording):
        imgui.text(rec.source_directory.stem or "Unknown")
        if rec.source_directory and imgui.is_item_hovered():
            gui_utils.draw_tooltip(str(rec.source_directory))

    def sort_and_filter_recordings(self, sort_specs_in: imgui.TableSortSpecs):
        if sort_specs_in.specs_dirty or self.require_sort:
            ids = list(self.recordings.keys())
            sort_specs = [sort_specs_in.get_specs(i) for i in range(sort_specs_in.specs_count)]
            for sort_spec in reversed(sort_specs):
                key = self._columns[sort_spec.column_index].sort_key_func
                ids.sort(key=key, reverse=sort_spec.get_sort_direction()==imgui.SortDirection.descending)
            self.sorted_recordings_ids = ids
            for flt in self.filters:
                key = lambda iid: flt.invert != flt.fun(iid, self.recordings[iid])
                if key is not None:
                    self.sorted_recordings_ids = list(filter(key, self.sorted_recordings_ids))
            if self.filter_box_text:
                search = self.filter_box_text.lower()
                def key(iid):
                    rec = self.get_rec_fun(self.recordings[iid])
                    if isinstance(rec, recording.Recording):
                        return \
                            search in rec.eye_tracker.value.lower() or \
                            search in rec.name.lower() or \
                            search in rec.participant.lower() or \
                            search in rec.project.lower()
                    else:
                        return \
                            search in 'camera' or \
                            search in rec.name.lower() or \
                            search in rec.video_file.lower()
                self.sorted_recordings_ids = list(filter(key, self.sorted_recordings_ids))
            sort_specs_in.specs_dirty = False
            self.require_sort = False


import pathlib
import mimetypes
import asyncio
import aiopath
import string
import pathvalidate
from dataclasses import dataclass, field
import datetime

from . import platform

@dataclass
class DirEntry:
    name: str
    is_dir: bool
    full_path: pathlib.Path
    ctime: float
    mtime: float
    size: int
    mime_type: str
    extra: dict = field(default_factory=dict)

    def __post_init__(self):
        if self.ctime is not None and not isinstance(self.ctime, datetime.datetime):
            self.ctime = datetime.datetime.fromtimestamp(self.ctime)
        if self.mtime is not None and not isinstance(self.mtime, datetime.datetime):
            self.mtime = datetime.datetime.fromtimestamp(self.mtime)


if platform.os==platform.Os.Windows:
    import ctypes
    _kernel32 = ctypes.WinDLL('kernel32',use_last_error=True)
    _mpr = ctypes.WinDLL('mpr',use_last_error=True)
    _netapi32 = ctypes.WinDLL('netapi32',use_last_error=True)
    _shell32 = ctypes.WinDLL('shell32',use_last_error=True)


    ERROR_NO_MORE_ITEMS = 259
    ERROR_EXTENDED_ERROR= 1208
    ERROR_SESSION_CREDENTIAL_CONFLICT = 1219

    def _error_check_non0_is_error_ex(allowed, result, func, args):
        if not result or result in allowed:
            return result
        if result==ERROR_EXTENDED_ERROR:
            last_error = ctypes.wintypes.DWORD()
            provider = ctypes.create_unicode_buffer(256)
            description = ctypes.create_unicode_buffer(256)
            WNetGetLastError(ctypes.byref(last_error),description,ctypes.sizeof(description),provider,ctypes.sizeof(provider))
            raise ctypes.WinError(last_error.value, f'{provider.value} failed: {description.value}')
        if (err := ctypes.get_last_error()):
            result = err
        raise ctypes.WinError(result)
    def _error_check_non0_is_error(result, func, args):
        return _error_check_non0_is_error_ex([], result, func, args)
    def _error_check_0_is_error(result, func, args):
        if not result:
            raise ctypes.WinError(ctypes.get_last_error())
        return result

    SHGFP_TYPE_CURRENT  = 0     # Get current, not default value
    CSIDL_DESKTOP       = 0
    CSIDL_PROFILE       = 40
    CSIDL_MYDOCUMENTS   = 5     # AKA CSIDL_PERSONAL
    SHGetFolderPath = _shell32.SHGetFolderPathW
    SHGetFolderPath.argtypes = ctypes.wintypes.HWND, ctypes.c_int, ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD, ctypes.wintypes.LPWSTR
    SHGetFolderPath.restype = ctypes.wintypes.DWORD
    SHGetFolderPath.errcheck = _error_check_non0_is_error

    class NETRESOURCE(ctypes.Structure):
        _fields_ = [
            ('scope', ctypes.wintypes.DWORD),
            ('type', ctypes.wintypes.DWORD),
            ('display_type', ctypes.wintypes.DWORD),
            ('usage', ctypes.wintypes.DWORD),
            ('local_name', ctypes.wintypes.LPWSTR),
            ('remote_name', ctypes.wintypes.LPWSTR),
            ('comment', ctypes.wintypes.LPWSTR),
            ('provider', ctypes.wintypes.LPWSTR),
        ]
    LPNETRESOURCE = ctypes.POINTER(NETRESOURCE)

    RESOURCE_CONNECTED  = 0x00000001
    RESOURCE_GLOBALNET  = 0x00000002
    RESOURCE_REMEMBERED = 0x00000003
    RESOURCE_RECENT     = 0x00000004
    RESOURCE_CONTEXT    = 0x00000005
    RESOURCETYPE_ANY        = 0x00000000
    RESOURCETYPE_DISK       = 0x00000001
    RESOURCETYPE_PRINT      = 0x00000002
    RESOURCETYPE_RESERVED   = 0x00000008
    RESOURCETYPE_UNKNOWN    = 0xFFFFFFFF
    RESOURCEUSAGE_CONNECTABLE   = 0x00000001
    RESOURCEUSAGE_CONTAINER     = 0x00000002
    RESOURCEUSAGE_NOLOCALDEVICE = 0x00000004
    RESOURCEUSAGE_SIBLING       = 0x00000008
    RESOURCEUSAGE_ATTACHED      = 0x00000010
    RESOURCEDISPLAYTYPE_GENERIC         = 0x00000000
    RESOURCEDISPLAYTYPE_DOMAIN          = 0x00000001
    RESOURCEDISPLAYTYPE_SERVER          = 0x00000002
    RESOURCEDISPLAYTYPE_SHARE           = 0x00000003
    RESOURCEDISPLAYTYPE_FILE            = 0x00000004
    RESOURCEDISPLAYTYPE_GROUP           = 0x00000005
    RESOURCEDISPLAYTYPE_NETWORK         = 0x00000006
    RESOURCEDISPLAYTYPE_ROOT            = 0x00000007
    RESOURCEDISPLAYTYPE_SHAREADMIN      = 0x00000008
    RESOURCEDISPLAYTYPE_DIRECTORY       = 0x00000009
    RESOURCEDISPLAYTYPE_TREE            = 0x0000000A
    RESOURCEDISPLAYTYPE_NDSCONTAINER    = 0x0000000B

    WNetGetLastError = _mpr.WNetGetLastErrorW
    WNetGetLastError.argtypes = ctypes.wintypes.LPDWORD, ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD, ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD
    WNetGetLastError.restype = ctypes.wintypes.DWORD
    WNetGetLastError.errcheck = _error_check_non0_is_error

    WNetOpenEnum = _mpr.WNetOpenEnumW
    WNetOpenEnum.argtypes = ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, LPNETRESOURCE, ctypes.wintypes.LPHANDLE
    WNetOpenEnum.restype = ctypes.wintypes.DWORD
    WNetOpenEnum.errcheck = _error_check_non0_is_error

    WNetEnumResource = _mpr.WNetEnumResourceW
    WNetEnumResource.argtypes = ctypes.wintypes.HANDLE, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPVOID, ctypes.wintypes.LPDWORD
    WNetEnumResource.restype = ctypes.wintypes.DWORD
    WNetEnumResource.errcheck = lambda r,f,a: _error_check_non0_is_error_ex([ERROR_NO_MORE_ITEMS], r, f, a)

    CONNECT_TEMPORARY = 0x00000004
    WNetAddConnection2 = _mpr.WNetAddConnection2W
    WNetAddConnection2.argtypes = LPNETRESOURCE, ctypes.wintypes.LPCWSTR, ctypes.wintypes.LPCWSTR, ctypes.wintypes.DWORD
    WNetAddConnection2.restype = ctypes.wintypes.DWORD
    WNetAddConnection2.errcheck = lambda r,f,a: _error_check_non0_is_error_ex([ERROR_SESSION_CREDENTIAL_CONFLICT], r, f, a)

    WNetGetUser = _mpr.WNetGetUserW
    WNetGetUser.argtypes = ctypes.wintypes.LPCWSTR, ctypes.wintypes.LPWSTR, ctypes.wintypes.LPDWORD
    WNetGetUser.restype = ctypes.wintypes.DWORD
    WNetGetUser.errcheck = _error_check_non0_is_error

    WNetCancelConnection2 = _mpr.WNetCancelConnection2W
    WNetCancelConnection2.argtypes = ctypes.wintypes.LPCWSTR, ctypes.wintypes.DWORD, ctypes.wintypes.BOOL
    WNetCancelConnection2.restype = ctypes.wintypes.DWORD
    WNetCancelConnection2.errcheck = _error_check_non0_is_error

    WNetCloseEnum = _mpr.WNetCloseEnum
    WNetCloseEnum.argtypes = ctypes.wintypes.HANDLE,
    WNetCloseEnum.restype = ctypes.wintypes.DWORD
    WNetCloseEnum.errcheck = _error_check_non0_is_error


    GMEM_FIXED    = 0x0000
    GMEM_MOVEABLE = 0x0002
    GMEM_ZEROINIT = 0x0040
    GHND          = GMEM_ZEROINIT | GMEM_MOVEABLE
    GPTR          = GMEM_ZEROINIT | GMEM_FIXED
    GlobalAlloc = _kernel32.GlobalAlloc
    GlobalAlloc.argtypes = ctypes.wintypes.UINT, ctypes.c_ssize_t
    GlobalAlloc.restype = ctypes.wintypes.HANDLE
    GlobalFree = _kernel32.GlobalFree
    GlobalFree.argtypes = ctypes.wintypes.HANDLE,
    GlobalFree.restype = ctypes.wintypes.HANDLE
    GlobalFree.errcheck = _error_check_non0_is_error


    class UNIVERSAL_NAME_INFO(ctypes.Structure):
        _fields_ = [
            ('universal_name', ctypes.wintypes.LPWSTR),
        ]
    LPUNIVERSAL_NAME_INFO = ctypes.POINTER(UNIVERSAL_NAME_INFO)
    class REMOTE_NAME_INFO(ctypes.Structure):
        _fields_ = [
            ('universal_name', ctypes.wintypes.LPWSTR),
            ('connection_name', ctypes.wintypes.LPWSTR),
            ('remaining_path', ctypes.wintypes.LPWSTR),
        ]
    LPREMOTE_NAME_INFO = ctypes.POINTER(REMOTE_NAME_INFO)
    UNIVERSAL_NAME_INFO_LEVEL   = 0x00000001
    REMOTE_NAME_INFO_LEVEL      = 0x00000002
    WNetGetUniversalName = _mpr.WNetGetUniversalNameW
    WNetGetUniversalName.argtypes = ctypes.wintypes.LPCWSTR, ctypes.wintypes.DWORD, ctypes.wintypes.LPVOID, ctypes.wintypes.LPDWORD
    WNetGetUniversalName.restype = ctypes.wintypes.DWORD
    WNetGetUniversalName.errcheck = _error_check_non0_is_error

    GetLogicalDrives = _kernel32.GetLogicalDrives
    GetLogicalDrives.argtypes = None
    GetLogicalDrives.restype = ctypes.wintypes.DWORD
    GetLogicalDrives.errcheck = _error_check_0_is_error

    GetVolumeInformation = _kernel32.GetVolumeInformationW
    GetVolumeInformation.argtypes = ctypes.wintypes.LPCWSTR, ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD
    GetVolumeInformation.restype = ctypes.wintypes.BOOL
    GetVolumeInformation.errcheck = _error_check_0_is_error

    GetDriveType = _kernel32.GetDriveTypeW
    GetDriveType.argtypes = ctypes.wintypes.LPCWSTR,
    GetDriveType.restype = ctypes.wintypes.UINT

    GetDiskFreeSpaceEx = _kernel32.GetDiskFreeSpaceExW
    GetDiskFreeSpaceEx.argtypes = ctypes.wintypes.LPCWSTR, ctypes.wintypes.PULARGE_INTEGER, ctypes.wintypes.PULARGE_INTEGER, ctypes.wintypes.PULARGE_INTEGER
    GetDiskFreeSpaceEx.restype = ctypes.wintypes.BOOL
    GetDiskFreeSpaceEx.errcheck = _error_check_0_is_error


    LMCSTR = ctypes.wintypes.LPCWSTR
    LMSTR = ctypes.wintypes.LPWSTR
    NET_API_STATUS = ctypes.wintypes.DWORD
    NERR_Success = 0
    NERR_NetNameNotFound = 2100+210
    MAX_PREFERRED_LENGTH = ctypes.wintypes.DWORD(-1)

    class SHARE_INFO_0(ctypes.Structure):
        _fields_ = [
            ('netname', LMSTR),
        ]
    LPSHARE_INFO_0 = ctypes.POINTER(SHARE_INFO_0)
    class SHARE_INFO_1(ctypes.Structure):
        _fields_ = [
            ('netname', LMSTR),
            ('type', ctypes.wintypes.DWORD),
            ('remark', LMSTR),
        ]
    LPSHARE_INFO_1 = ctypes.POINTER(SHARE_INFO_1)

    # SHARE_INFO_1.type flags
    STYPE_DISKTREE          = 0
    STYPE_PRINTQ            = 1
    STYPE_DEVICE            = 2
    STYPE_IPC               = 3
    STYPE_TEMPORARY         = 0x40000000
    STYPE_SPECIAL           = 0x80000000

    NetShareEnum = _netapi32.NetShareEnum
    NetShareEnum.argtypes = (LMSTR, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.LPBYTE), ctypes.wintypes.DWORD, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPDWORD, ctypes.wintypes.LPDWORD)
    NetShareEnum.restype = NET_API_STATUS
    NetShareEnum.errcheck = _error_check_non0_is_error

    NetShareGetInfo = _netapi32.NetShareGetInfo
    NetShareGetInfo.argtypes = (LMSTR, LMSTR, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.LPBYTE))
    NetShareGetInfo.restype = NET_API_STATUS
    NetShareGetInfo.errcheck = lambda r,f,a: _error_check_non0_is_error_ex([NERR_NetNameNotFound], r, f, a)

    NetApiBufferFree = _netapi32.NetApiBufferFree
    NetApiBufferFree.argtypes = (ctypes.wintypes.LPVOID,)
    NetApiBufferFree.restype = NET_API_STATUS



    def get_thispc_listing() -> list[DirEntry]:
        items = []
        # NB: we also check for Downloads folder (doesn't have a CSIDL), add if found
        # expect it as a subfolder of user's profile folder
        for d, subdir, disp_name in [(CSIDL_DESKTOP, None, 'Desktop'), (CSIDL_MYDOCUMENTS, None, 'My Documents'), (CSIDL_PROFILE, 'Downloads', 'Downloads')]:
            buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
            SHGetFolderPath(None, d, None, SHGFP_TYPE_CURRENT, buf)
            path = pathlib.Path(buf.value)
            if subdir:
                path /= subdir
                if not path.is_dir():
                    continue
            stat = path.stat()
            item = DirEntry(disp_name, path.is_dir(), path,
                            stat.st_ctime, stat.st_mtime, stat.st_size,
                            mimetypes.guess_type(path)[0])
            items.append(item)
        return items


    def get_drives() -> list[DirEntry]:
        drives = []
        bitmask = GetLogicalDrives()
        for letter in string.ascii_uppercase:
            if bitmask & 1:
                drive = f'{letter}:\\'
                entry = DirEntry(drive[0:-1],True,pathlib.Path(drive),None,None,None,None)
                # get name of drive
                drive_name = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH+1)
                if not GetVolumeInformation(drive,drive_name,ctypes.sizeof(drive_name),None,None,None,None,0):
                    raise ctypes.WinError(ctypes.get_last_error())
                entry.extra['drive_name'] = drive_name.value
                # get drive type
                match GetDriveType(drive):
                    case 0 | 1:     # DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR
                        # we skip these
                        continue
                    case 2:     # DRIVE_REMOVABLE
                        # assume a USB drive, by far most likely. getting what it truly is seems complicated
                        entry.mime_type = 'file_action/drive_removable'
                        display_name = drive_name.value if drive_name.value else 'USB Drive'
                        entry.name = f'{display_name} ({drive[0:-1]})'
                    case 3:     # DRIVE_FIXED
                        entry.mime_type = 'file_action/drive'
                        display_name = drive_name.value if drive_name.value else 'Local Disk'
                        entry.name = f'{display_name} ({drive[0:-1]})'
                    case 4:     # DRIVE_REMOTE
                        entry.mime_type = 'file_action/drive_network'
                        buffer_len = ctypes.wintypes.DWORD(1024)
                        buffer = ctypes.create_unicode_buffer(buffer_len.value)
                        un_buffer = ctypes.cast(buffer,LPUNIVERSAL_NAME_INFO)
                        WNetGetUniversalName(drive, UNIVERSAL_NAME_INFO_LEVEL, un_buffer, ctypes.byref(buffer_len))
                        entry.extra['remote_path'] = un_buffer[0].universal_name
                        # build name: 'share_name (net_name) (drive)'
                        path_comps = split_network_path(entry.extra['remote_path'])
                        if len(path_comps)>=2:
                            entry.name = f'{path_comps[1]} ({path_comps[0]}) ({drive[0:-1]})'
                    case 5:     # DRIVE_CDROM
                        entry.mime_type = 'file_action/drive_cdrom'
                    case 6:     # DRIVE_RAMDISK
                        entry.mime_type = 'file_action/drive_ramdisk'
                # get size information
                total, free = ctypes.wintypes.ULARGE_INTEGER(), ctypes.wintypes.ULARGE_INTEGER()
                GetDiskFreeSpaceEx(drive,None,ctypes.byref(total),ctypes.byref(free))
                entry.size = total.value
                entry.extra['free_space'] = free.value

                drives.append(entry)
            bitmask >>= 1

        return drives

    def get_visible_shares(server: str, user: str='', password: str='', domain=''):
        shares: list[DirEntry] = []

        # if no user provided, user should be able to connect from the Windows
        # explorer to this server (e.g. user should have provided credentials)
        # if you can open it in explorer, you can list it using this call)
        if user:
            need_conn_cleanup = _server_login(server, user, password, domain)
        else:
            need_conn_cleanup = False

        server = server.strip('\\/')
        nr = NETRESOURCE(remote_name=f'\\\\{server}')

        hEnum = ctypes.wintypes.HANDLE()
        WNetOpenEnum(RESOURCE_GLOBALNET, RESOURCETYPE_ANY, 0, nr, ctypes.byref(hEnum))

        # allocate a buffer
        cbBuffer = ctypes.wintypes.DWORD(16384)     # 16K is a good size
        mem_handle = GlobalAlloc(GPTR, cbBuffer.value)
        lpnrLocal = ctypes.cast(mem_handle,LPNETRESOURCE)

        while True:
            cEntries = ctypes.wintypes.DWORD(-1)        # enumerate all possible entries
            result = WNetEnumResource(hEnum, ctypes.byref(cEntries),lpnrLocal,ctypes.byref(cbBuffer))
            if result==ERROR_NO_MORE_ITEMS:
                break

            for idx in range(cEntries.value):
                share = lpnrLocal[idx]
                share_path = share.remote_name
                path_comps = split_network_path(share_path)
                if len(path_comps)>=2:
                    shares.append(DirEntry(path_comps[1],True,pathlib.Path(share_path),None,None,None,'file_action/net_share'))

        # clean up
        GlobalFree(mem_handle)
        WNetCloseEnum(hEnum)
        if need_conn_cleanup:
            _server_logout(server)

        return shares

    def get_all_shares(server: str, user: str='', password: str='', domain=''):
        # similar to get_visible_shares() but also lists hidden shares
        shares: list[DirEntry] = []

        # if no user provided, user should be able to connect from the Windows
        # explorer to this server (e.g. user should have provided credentials)
        # if you can open it in explorer, you can list it using this call)
        if user:
            need_conn_cleanup = _server_login(server, user, password, domain)
        else:
            need_conn_cleanup = False

        server = server.strip('\\/')
        buf = ctypes.wintypes.LPBYTE()
        buf_len = MAX_PREFERRED_LENGTH
        read, total = ctypes.wintypes.DWORD(0), ctypes.wintypes.DWORD(0)
        NetShareEnum(f'\\\\{server}', 1, ctypes.byref(buf), buf_len, ctypes.byref(read), ctypes.byref(total), None)

        UserInfoArray = SHARE_INFO_1 * read.value
        remote_shares = UserInfoArray.from_address(ctypes.addressof(buf.contents))
        for share in remote_shares:
            shares.append(DirEntry(share.netname,True,pathlib.Path(f'\\\\{server}\\{share.netname}'),None,None,None,'file_action/net_share'))

        # clean up
        NetApiBufferFree(buf)
        if need_conn_cleanup:
            _server_logout(server)

        return shares

    def check_share(server: str, share: str, user: str='', password: str='', domain=''):
        # if no user provided, user should be able to connect from the Windows
        # explorer to this server (e.g. user should have provided credentials)
        # if you can open it in explorer, you can list it using this call)
        if user:
            need_conn_cleanup = _server_login(server, user, password, domain)
        else:
            need_conn_cleanup = False

        server = server.strip('\\/')
        buf = ctypes.wintypes.LPBYTE()
        ret = NetShareGetInfo(server,share,0,buf)   # NB: I never got NetShareCheck to work, always return that share doesn't exist, so using this slightly round-about way (we can't get info of a non-existent share)
        if ret==NERR_Success:
            NetApiBufferFree(buf)

        if need_conn_cleanup:
            _server_logout(server)

        return ret!=NERR_NetNameNotFound

    def _server_login(server: str, user: str, password: str, domain=''):
        # check if we already have a connection with the server, and if so, if
        # it is the expected user
        user_name = ctypes.create_unicode_buffer(1024)
        buf_size = ctypes.wintypes.DWORD(ctypes.sizeof(user_name))
        server = server.strip('\\/')
        server = f'\\\\{server}'
        try:
            WNetGetUser(server, user_name, ctypes.byref(buf_size))
        except:
            need_conn = True
        else:
            user_name = user_name.value
            need_conn = user_name!=user and not user_name.endswith(user)
        if need_conn:
            nr = NETRESOURCE(type=RESOURCETYPE_DISK, remote_name=server)
            if domain:
                user = f'{domain}\\{user}'
            res = WNetAddConnection2(nr, password, user, CONNECT_TEMPORARY)
            if res==ERROR_SESSION_CREDENTIAL_CONFLICT:
                # apparently we are connected even if WNetGetUser() doesn't know. Carry on and hope for the best
                need_conn = False
            elif res!=0:
                _error_check_non0_is_error(res,None,None)
        return need_conn
    def _server_logout(server: str):
        server = server.strip('\\/')
        server = f'\\\\{server}'
        try:
            WNetCancelConnection2(server, 0, True)
        except:
            # ok if this fails, we did our best
            pass

    def split_network_path(path: str|pathlib.Path) -> list[str]:
        path = str(path)
        if not path.startswith(('\\\\','//')):
            return []
        # this is a network address
        # split into components
        path = path.strip('\\/').replace('\\','/')
        return [s for s in str(path).split('/') if s]

    def get_net_computer(path: str|pathlib.Path):
        # determine if it is a network computer (\\SERVER) and not a path including share (\\SERVER\share)
        net_comp = split_network_path(path)
        if len(net_comp)==1:    # a single name entry, so thats just a computer
            return net_comp[0]
        return None


async def get_dir_list(path: pathlib.Path) -> list[DirEntry]:
    # will throw when path doesn't exist or is not a directory
    path = aiopath.AsyncPath(path)
    out = []
    async for e in path.iterdir():
        try:
            stat, is_dir = await asyncio.gather(e.stat(), e.is_dir())
            item = DirEntry(e.name, is_dir, pathlib.Path(e),
                            stat.st_ctime, stat.st_mtime, stat.st_size,
                            mimetypes.guess_type(e)[0])
        except:
            pass
        else:
            out.append(item)

    return out

def get_dir_list_sync(path: pathlib.Path) -> list[DirEntry]:
    # will throw when path doesn't exist or is not a directory
    path = pathlib.Path(path)
    out = []
    for e in path.iterdir():
        try:
            stat = e.stat()
            item = DirEntry(e.name, e.is_dir(), pathlib.Path(e),
                            stat.st_ctime, stat.st_mtime, stat.st_size,
                            mimetypes.guess_type(e)[0])
        except:
            pass
        else:
            out.append(item)

    return out

async def make_dir(path: str | pathlib.Path, exist_ok: bool = False):
    pathvalidate.validate_filepath(path, "auto")
    path = aiopath.AsyncPath(path)
    await path.mkdir(exist_ok=exist_ok)

async def rename_path(old_path: str | pathlib.Path, new_path: str | pathlib.Path):
    pathvalidate.validate_filepath(old_path, "auto")
    pathvalidate.validate_filepath(new_path, "auto")
    return await aiopath.AsyncPath(old_path).rename(new_path)



import math
import numpy as np
from typing import Any
from matplotlib import ticker
from imgui_bundle import imgui, hello_imgui, icons_fontawesome_6 as ifa6

from .. import annotation, timestamps, utils

def _color_u32_replace_alpha(color: int, alpha: float) -> int:
    col = imgui.get_style_color_vec4(color)
    return imgui.color_convert_float4_to_u32((col.x, col.y, col.z, alpha))

def color_replace_alpha(color: imgui.ImColor, alpha: float) -> imgui.ImColor:
    return imgui.ImColor(color.value.x, color.value.y, color.value.z, alpha)

def _color_luminance(color: imgui.ImColor) -> float:
    return (0.2126 * color.value.x) + (0.7152 * color.value.y) + (0.0722 * color.value.z)

def _calc_contrast_ratio(a: imgui.ImColor, b: imgui.ImColor) -> float:
    y1 = _color_luminance(a)
    y2 = _color_luminance(b)
    return (y2 + 0.05) / (y1 + 0.05) if (y1 > y2) else (y1 + 0.05) / (y2 + 0.05)

def color_adjust_contrast(color: imgui.ImColor, contrast: float, bg_color: imgui.ImColor) -> imgui.ImColor:
    bg_contrast_ratio = _calc_contrast_ratio(bg_color, color)
    return imgui.ImColor(0, 0, 0, 1 - bg_contrast_ratio * 0.45) if (bg_contrast_ratio > contrast) else color

def color_brighten(color: imgui.ImColor, amount: float):
    amount = 1. / (1. + amount)
    return imgui.ImColor(1. - (1. - color.value.x) * amount, 1. - (1. - color.value.y) * amount,
                         1. - (1. - color.value.z) * amount, color.value.w)

def color_darken(color: imgui.ImColor, amount: float):
    amount = 1. / (1. + amount)
    return imgui.ImColor(color.value.x * amount, color.value.y * amount, color.value.z * amount)


class Timeline:
    def __init__(self, video_ts: timestamps.VideoTimestamps, annotations: dict[annotation.Event, list[int]] = None):
        self._video_ts = video_ts
        self._duration = self._video_ts.get_last()[1]/1000. # ms -> s

        # horizontal scroll position, in pixels
        self._h_scroll      = 0
        self._scale         = 1.
        self._new_h_scroll  = None

        self._current_time = 0.
        self._dragged_time: float = None
        self._view_scale_fac: float = None

        # axis ticker
        self._major_ticker = ticker.MaxNLocator()
        self._major_ticks_pos: np.ndarray
        self._n_minor_divs: int
        self._minor_ticks_pos: np.ndarray
        self._major_ticks_lbl: list[str]

        # information about GUI size
        self.draw_width: int = None

        # GUI interaction possibilities
        self._allow_seek = False
        self._allow_annotate = False
        self._allow_timeline_zoom = False

        # tracks
        self._annotations_frame = annotations
        self._annotations       = self._annotations_to_time()
        self._annotation_colors = self._make_annotation_colors()
        self._show_annotation_labels = True
        self._show_info_on_hover = True
        self._annotation_tooltips: dict[annotation.Event, str] = {}

        # communication with owner
        self._requests: list[tuple[str,Any]] = []


    def get_fixed_elements_height(self) -> float:
        timeline_height = imgui.get_font_size() + 2*imgui.get_style().frame_padding.y
        return timeline_height + (imgui.get_style().scrollbar_size if self._allow_timeline_zoom else 0)

    def get_num_annotations(self) -> int:
        return 0 if self._annotations_frame is None else len(self._annotations_frame)

    def set_allow_seek(self, allow_seek: bool):
        self._allow_seek = allow_seek

    def set_allow_annotate(self, allow_annotate: bool):
        self._allow_annotate = allow_annotate

    def set_allow_timeline_zoom(self, allow_timeline_zoom: bool):
        self._allow_timeline_zoom = allow_timeline_zoom

    def set_show_annotation_labels(self, show_label):
        self._show_annotation_labels = show_label

    def set_show_info_on_hover(self, show_info_on_hover: bool):
        self._show_info_on_hover = show_info_on_hover

    def set_annotation_keys(self, annotate_shortcut_key_map: dict[annotation.Event, imgui.Key], annotate_tooltips: dict[annotation.Event, str] = None):
        self._annotation_tooltips.clear()
        for e in annotation.Event:
            tool_tip = ''
            if e in annotate_tooltips:
                tool_tip = annotate_tooltips[e]
                if e in annotate_shortcut_key_map and self._allow_annotate:
                    tool_tip += f' ({imgui.get_key_name(annotate_shortcut_key_map[e])})'
            elif e in annotate_shortcut_key_map and self._allow_annotate:
                tool_tip = imgui.get_key_name(annotate_shortcut_key_map[e])

            if tool_tip:
                self._annotation_tooltips[e] = tool_tip

    def _annotations_to_time(self) -> dict[annotation.Event, list[float]]:
        if not self._annotations_frame:
            return None
        return {e:[self._video_ts.get_timestamp(i)/1000 for i in self._annotations_frame[e]] for e in self._annotations_frame}  # ms -> s

    def _make_annotation_colors(self) -> dict[annotation.Event, imgui.ImColor]:
        if not self._annotations_frame:
            return {}

        colors = utils.get_colors(len(self._annotations_frame), 0.45, 0.65)
        return {k:imgui.ImColor(*c) for k,c in zip(self._annotations_frame, colors)}

    def get_annotation_colors(self) -> dict[annotation.Event, imgui.ImColor]:
        return self._annotation_colors

    def set_position(self, time: float, frame_idx: int):
        self._current_time = max(0,min(time,self._duration))
        self._current_frame= frame_idx

    def get_requests(self) -> list[tuple[str,Any]]:
        reqs = self._requests
        self._requests = []
        return reqs

    def notify_annotations_changed(self):
        self._annotations = self._annotations_to_time()

    def _request_time(self, time: float):
        self._requests.append(('seek', time))

    def _request_delete(self, event: annotation.Event, frame_idxs: list[int]):
        if not isinstance(frame_idxs,list):
            frame_idxs = [frame_idxs]
        self._requests.append(('delete_coding', (event,frame_idxs)))

    def _calc_scale_fac(self):
        # yields s/pix on the timeline
        self._view_scale_fac = self._duration / self.draw_width

    def _pos_to_time(self, pos_x, is_window_pos) -> float:
        if is_window_pos:
            pos_x -= self.content_origin.x
        # position as fraction of self.duration
        duration_frac = pos_x * self._view_scale_fac / self._duration + self._h_scroll/self.draw_width
        # in seconds
        return duration_frac * self._duration

    def _time_to_pos(self, time: float, as_window_pos) -> float:
        pos = time / self._view_scale_fac
        if as_window_pos:
            pos += imgui.get_cursor_screen_pos().x
        return pos

    def _determine_ticks(self):
        # calculate number of major ticks
        # how big is a tick label?
        width = imgui.calc_text_size('0:00:00.000').x
        num_ticks = math.floor(self.draw_width/width/2)

        # generate ticks
        self._major_ticker.set_params(nbins=num_ticks)
        major_ticks_time = np.array([t for t in self._major_ticker.tick_values(0, self._duration) if t>=0 and t<=self._duration])

        # generate minor ticks, logic similar to matplotlib.ticker.AutoMinorLocator
        majorstep = major_ticks_time[1] - major_ticks_time[0]
        majorstep_mantissa = 10 ** (np.log10(majorstep) % 1)
        self._n_minor_divs = 5 if np.isclose(majorstep_mantissa, [1, 2.5, 5, 10]).any() else 4
        minorstep = majorstep / self._n_minor_divs
        tmax = round(self._duration / minorstep) + 1
        minor_ticks_time = (np.arange(0, tmax) * minorstep)

        # convert from time to pixels
        self._major_ticks_pos = self._time_to_pos(major_ticks_time, False)
        self._minor_ticks_pos = self._time_to_pos(minor_ticks_time, False)

        # format the values
        # determine if we have a ms part
        need_ms = (majorstep%1) > np.finfo(np.dtype(float)).eps*10
        self._major_ticks_lbl = [utils.format_duration(x, need_ms) for x in major_ticks_time]

    def _draw_time_ruler(self):
        style = imgui.get_style()
        dpi_fac = hello_imgui.dpi_window_size_factor()

        # get input state we'll need
        drag_delta = imgui.get_mouse_drag_delta(imgui.MouseButton_.left)
        mouse_pos = imgui.get_mouse_pos()
        cursor_pos = imgui.get_cursor_screen_pos()
        size = imgui.ImVec2(self.draw_width, imgui.get_font_size()+2*style.frame_padding.y)

        # set up for interaction with the timeline
        if self._allow_seek or self._allow_timeline_zoom:
            # 1. make button over whole width
            imgui.invisible_button('##play_head_control', size)
        else:
            imgui.dummy(size)

        # enable moving play head
        if self._allow_seek:
            # 2. if pressed on the button, move play head
            if imgui.is_item_activated() or (imgui.is_item_active() and abs(drag_delta.x)>.001):
                self._request_time(self._pos_to_time(mouse_pos.x, True))
                imgui.reset_mouse_drag_delta()  # mark that drag has been processed
        if self._allow_timeline_zoom:
            self._handle_zoom(imgui.is_item_hovered())

        # draw ticks
        draw_list = imgui.get_window_draw_list()
        draw_list.push_clip_rect(cursor_pos, cursor_pos+size)

        lbl_color   = imgui.get_color_u32(imgui.Col_.text)
        tick_color  = imgui.get_color_u32(imgui.Col_.separator)

        tick_pos_y = cursor_pos.y + size.y
        for lbl,pos in zip(self._major_ticks_lbl,self._major_ticks_pos):
            rounded_gridline_pos_x = round(cursor_pos.x + pos)
            draw_list.add_text((rounded_gridline_pos_x + 4*dpi_fac, cursor_pos.y + 2*style.frame_padding.y - 2*dpi_fac), lbl_color, lbl)
            draw_list.add_line((rounded_gridline_pos_x, tick_pos_y - 10*dpi_fac), (rounded_gridline_pos_x, tick_pos_y - 2*dpi_fac), tick_color, thickness=dpi_fac)

        # draw play head
        playhead_screen_position = self._time_to_pos(self._dragged_time if self._dragged_time is not None else self._current_time, True)
        if self.get_num_annotations():
            playhead_screen_position = playhead_screen_position - size.y * 0.5
            draw_list.add_triangle_filled(
                (playhead_screen_position           , cursor_pos.y + 2.5*dpi_fac),
                (playhead_screen_position + size.y  , cursor_pos.y + 2.5*dpi_fac),
                (playhead_screen_position + size.y/2, cursor_pos.y + size.y - 2.5*dpi_fac),
                0xE553A3F9)
        else:
            playhead_line_pos = imgui.ImVec2(playhead_screen_position, cursor_pos.y)
            draw_list.add_line(playhead_line_pos,
                               playhead_line_pos + imgui.ImVec2(0, size.y),
                               0xE553A3F9,
                               thickness=2*dpi_fac)

        # draw vertical separator
        curr_pos = imgui.get_cursor_screen_pos()
        draw_list.add_line(
            (curr_pos.x, curr_pos.y - 1*dpi_fac),
            (curr_pos.x + self.draw_width, curr_pos.y - 1*dpi_fac),
            imgui.get_color_u32(imgui.Col_.separator))

        draw_list.pop_clip_rect()

    def _handle_zoom(self, is_hovered: bool):
        mouse_pos = imgui.get_mouse_pos()
        mouse_wheel = imgui.get_io().mouse_wheel
        if is_hovered and mouse_wheel!=0.:
            # clamp scaling, scale should not get smaller than 1
            delta_scale = .25*mouse_wheel
            delta_scale = max(1., self._scale+delta_scale)-self._scale
            if np.isclose(0., delta_scale):
                return

            self._scale += delta_scale
            new_h_scroll = min(max(0., self._h_scroll + mouse_pos.x*delta_scale), self.draw_width)
            if not np.isclose(self._h_scroll, new_h_scroll):
                self._new_h_scroll = new_h_scroll

    def _timepoint_interaction_logic(self, lbl: str, x_pos: float, y_ext: float, hover_info: tuple[str,float,int]=None, draggable=False, has_context_menu=True, add_episode_action=False) -> tuple[float, bool, str|None]:
        do_move = False
        action = None
        dragging = False

        if self._allow_seek or self._allow_annotate or has_context_menu or hover_info:
            dpi_fac = hello_imgui.dpi_window_size_factor()
            cursor_pos = imgui.get_cursor_screen_pos()
            imgui.set_cursor_screen_pos((x_pos-2*dpi_fac, cursor_pos.y))
            size = imgui.ImVec2(4*dpi_fac, y_ext)
            imgui.invisible_button(f'##timepoint_interactable_{lbl}', size)
            imgui.set_cursor_screen_pos(cursor_pos)

            is_active = imgui.is_item_active()
            is_deactivated = imgui.is_item_deactivated()
            is_hovered = imgui.is_item_hovered()

            if (is_active or is_hovered):
                if draggable and self._allow_seek:
                    imgui.set_mouse_cursor(imgui.MouseCursor_.resize_ew)
                elif self._allow_seek or self._allow_annotate:
                    imgui.set_mouse_cursor(imgui.MouseCursor_.hand)

            if hover_info and imgui.is_item_hovered(imgui.HoveredFlags_.for_tooltip | imgui.HoveredFlags_.delay_normal):
                imgui.set_tooltip(f'{hover_info[0]} {hover_info[1]}\ntimestamp: {hover_info[2]:.3f}\nframe index: {hover_info[3]}')

            drag_finished = False
            dragging = imgui.is_mouse_dragging(imgui.MouseButton_.left)
            if self._allow_seek and draggable:
                if dragging and is_active:
                    # start or update drag
                    x_pos = imgui.get_mouse_pos().x
                    self._dragged_time = self._pos_to_time(x_pos, True)
                else:
                    imgui.reset_mouse_drag_delta()  # mark that drag has been processed
                    drag_finished = self._dragged_time is not None
                    x_pos = imgui.get_mouse_pos().x
                    self._dragged_time = None
            do_move = self._allow_seek and (is_deactivated or (draggable and drag_finished))

            if has_context_menu and (self._allow_seek or self._allow_annotate) and imgui.begin_popup_context_item(f"##timepoint_context_menu_{lbl}"):
                if self._allow_seek and imgui.selectable(ifa6.ICON_FA_ARROW_RIGHT + ' go to timepoint', False)[0]:
                    do_move = True
                if self._allow_annotate and imgui.selectable(ifa6.ICON_FA_TRASH_CAN + ' remove timepoint', False)[0]:
                    action = 'delete_timepoint'
                if self._allow_annotate and add_episode_action and imgui.selectable(ifa6.ICON_FA_TRASH_CAN + ' remove episode', False)[0]:
                    action = 'delete_episode'
                imgui.end_popup()

        return x_pos, do_move, action

    def _episode_interaction_logic(self, lbl: str, x_start: float, x_end: float, y_ext: float, hover_info: tuple[str,int,list[float],list[int]]=None) -> tuple[float, bool, str|None]:
        x_pos = x_start
        do_move = False
        action = None

        if self._allow_seek or self._allow_annotate or hover_info:
            dpi_fac = hello_imgui.dpi_window_size_factor()
            cursor_pos = imgui.get_cursor_screen_pos()
            imgui.set_cursor_screen_pos((x_start+2*dpi_fac, cursor_pos.y))
            size = imgui.ImVec2(x_end-x_start-4*dpi_fac, y_ext)
            imgui.invisible_button(f'##episode_interactable_{lbl}', size)
            imgui.set_cursor_screen_pos(cursor_pos)

            if hover_info and imgui.is_item_hovered(imgui.HoveredFlags_.for_tooltip | imgui.HoveredFlags_.delay_normal):
                imgui.set_tooltip(f'{hover_info[0]} {hover_info[1]}\ntimestamps: {hover_info[2][0]:.3f} - {hover_info[2][1]:.3f}\nframe indices: {hover_info[3][0]} - {hover_info[3][1]}')

            if (self._allow_seek or self._allow_annotate) and imgui.begin_popup_context_item(f"##episode_context_menu_{lbl}"):
                if self._allow_seek and imgui.selectable(ifa6.ICON_FA_ARROW_LEFT + ' go to start', False)[0]:
                    do_move = True
                if self._allow_seek and imgui.selectable(ifa6.ICON_FA_ARROW_RIGHT + ' go to end', False)[0]:
                    x_pos = x_end
                    do_move = True
                if self._allow_annotate and imgui.selectable(ifa6.ICON_FA_TRASH_CAN + ' remove episode', False)[0]:
                    action = 'delete_episode'
                imgui.end_popup()

        return x_pos, do_move, action

    def _draw_tracks(self):
        dpi_fac = hello_imgui.dpi_window_size_factor()
        cursor_pos = imgui.get_cursor_screen_pos()
        size       = imgui.get_content_region_avail()
        draw_list = imgui.get_window_draw_list()
        draw_list.push_clip_rect(cursor_pos, cursor_pos+size)

        # draw background, alternate color every division
        section_color = _color_u32_replace_alpha(imgui.Col_.separator, 0.12)
        for i in range(1,len(self._major_ticks_pos),2):
            start = self._major_ticks_pos[i]
            end   = self._major_ticks_pos[i+1] if i+1<len(self._major_ticks_pos) else self.draw_width
            draw_list.add_rect_filled(
                (cursor_pos.x+start, cursor_pos.y),
                (cursor_pos.x+end  , cursor_pos.y+size.y),
                section_color
            )

        # draw gridlines
        grid_color_major = _color_u32_replace_alpha(imgui.Col_.separator, 0.85)
        grid_color_minor = _color_u32_replace_alpha(imgui.Col_.separator, 0.3)
        for i,pos in enumerate(self._minor_ticks_pos):
            rounded_gridline_pos_x = round(cursor_pos.x + pos)
            if i%self._n_minor_divs==0:
                grid_color = grid_color_major
            else:
                grid_color = grid_color_minor
            draw_list.add_line((rounded_gridline_pos_x, cursor_pos.y), (rounded_gridline_pos_x, cursor_pos.y+size.y), grid_color, thickness=dpi_fac)

        # now draw the actual tracks
        height_per_track = size.y/len(self._annotations_frame)
        imgui.push_style_var(imgui.StyleVar_.item_spacing, (0,0))
        text_size = max([imgui.calc_text_size(e.value) for e in self._annotations_frame.keys()], key=lambda x: x.x)
        for e in self._annotations_frame:
            self._draw_track(e, self._annotations[e], self._annotations_frame[e], height_per_track, self._annotation_colors[e], text_size)
        imgui.pop_style_var()

        # draw player bar
        imgui.set_cursor_screen_pos(cursor_pos)
        playhead_screen_position = self._time_to_pos(self._dragged_time if self._dragged_time is not None else self._current_time, True)
        playhead_line_pos = imgui.ImVec2(playhead_screen_position, cursor_pos.y)
        draw_list.add_line(playhead_line_pos,
                           playhead_line_pos + imgui.ImVec2(0, size.y),
                           0xE553A3F9,
                           thickness=2*dpi_fac)
        new_pos, should_process, _ = self._timepoint_interaction_logic('player_bar', playhead_screen_position, size.y, None, draggable=True, has_context_menu=False)
        if should_process:
            self._request_time(self._pos_to_time(new_pos, True))

        # draw invisible button for interaction logic
        imgui.set_cursor_screen_pos(cursor_pos)
        imgui.invisible_button('##all_tracks', size)
        imgui.same_line(0,0)
        if self._allow_timeline_zoom:
            tracks_hovered = imgui.is_item_hovered()
            self._handle_zoom(tracks_hovered)

        draw_list.pop_clip_rect()

    def _draw_annotation_line(self, name: str, idx: int, time: float, frame_idx: int, seq_nr: int, height: float, border_color: int) -> str|None:
        dpi_fac = hello_imgui.dpi_window_size_factor()
        cursor_pos = imgui.get_cursor_screen_pos()
        draw_list = imgui.get_window_draw_list()
        screen_position = self._time_to_pos(time, True)
        draw_list.add_line((screen_position, cursor_pos.y),
                            (screen_position, cursor_pos.y+height),
                            border_color,
                            thickness=2*dpi_fac)
        hover_info = (name, seq_nr, time, frame_idx) if self._show_info_on_hover else None
        _, do_move, action = self._timepoint_interaction_logic(f'annotation_{name}_{idx}', screen_position, height, hover_info, draggable=False, has_context_menu=True)
        if do_move:
            self._request_time(self._pos_to_time(screen_position, True))
        return action

    def _draw_track(self, event: annotation.Event, annotations_time: list[float], annotations_frame: list[int], height: float, color: imgui.ImColor, text_size: imgui.ImVec2):
        dpi_fac = hello_imgui.dpi_window_size_factor()
        grid_color = _color_u32_replace_alpha(imgui.Col_.separator, 0.85)
        cursor_pos = imgui.get_cursor_screen_pos()
        draw_list = imgui.get_window_draw_list()
        size = imgui.ImVec2(imgui.get_content_region_avail().x, height)
        name = event.value
        draw_list.push_clip_rect(cursor_pos, cursor_pos+size)

        # draw separator below track
        y_pos = cursor_pos.y+size.y + 0.5*dpi_fac
        draw_list.add_line(
            (cursor_pos.x, y_pos),
            (cursor_pos.x + size.x, y_pos),
            grid_color, 2*dpi_fac)

        # setup for other drawing
        text_contrast_ratio = 1 / 1.57
        text_color = imgui.ImColor(imgui.get_style_color_vec4(imgui.Col_.text))
        text_color_adjusted = imgui.color_convert_float4_to_u32(color_adjust_contrast(text_color, text_contrast_ratio, color).value)
        bg_color     = imgui.color_convert_float4_to_u32(color_replace_alpha(color, color.value.w * 0.5).value)
        border_color = imgui.color_convert_float4_to_u32(color_replace_alpha(color, color.value.w * 0.9).value)

        text_offset = imgui.ImVec2((2*dpi_fac, 2*dpi_fac))
        clip_title_max = text_size + imgui.ImVec2([2*text_offset.x, 2*text_offset.y])

        # draw annotations
        if annotation.type_map[event]==annotation.Type.Interval:
            # intervals, draw as boxes
            action = None
            for m in range(0,len(annotations_time)-1,2):     # -1 to make sure we don't try incomplete intervals
                start_screen_position = self._time_to_pos(annotations_time[m],   True)
                end_screen_position   = self._time_to_pos(annotations_time[m+1], True)
                draw_list.add_rect_filled((start_screen_position, cursor_pos.y),
                                          (end_screen_position, cursor_pos.y+size.y),
                                          bg_color)
                draw_list.add_rect((start_screen_position, cursor_pos.y+dpi_fac/2),
                                   (end_screen_position, cursor_pos.y+size.y),
                                   border_color,
                                   thickness=dpi_fac)

                hover_info = (name, int(m/2)+1, annotations_time[m], annotations_frame[m]) if self._show_info_on_hover else None
                _, do_move, action = self._timepoint_interaction_logic(f'annotation_{name}_{m}', start_screen_position, size.y, hover_info, draggable=False, has_context_menu=True, add_episode_action=True)
                if do_move:
                    self._request_time(self._pos_to_time(start_screen_position, True))
                elif action=='delete_timepoint':
                    self._request_delete(event, annotations_frame[m])
                elif action=='delete_episode':
                    self._request_delete(event, annotations_frame[m:m+2])

                hover_info = (name, int(m/2)+1, annotations_time[m+1], annotations_frame[m+1]) if self._show_info_on_hover else None
                _, do_move, action = self._timepoint_interaction_logic(f'annotation_{name}_{m+1}', end_screen_position, size.y, hover_info, draggable=False, has_context_menu=True, add_episode_action=True)
                if do_move:
                    self._request_time(self._pos_to_time(end_screen_position, True))
                elif action=='delete_timepoint':
                    self._request_delete(event, annotations_frame[m+1])
                elif action=='delete_episode':
                    self._request_delete(event, annotations_frame[m:m+2])

                hover_info = (name, int(m/2)+1, annotations_time[m:m+2], annotations_frame[m:m+2]) if self._show_info_on_hover else None
                x_pos, do_move, action = self._episode_interaction_logic(f'annotation_{name}_{m}', start_screen_position, end_screen_position, size.y, hover_info)
                if do_move:
                    self._request_time(self._pos_to_time(x_pos, True))
                elif action=='delete_episode':
                    self._request_delete(event, annotations_frame[m:m+2])

            if len(annotations_time)%2:
                # unmatched marker left over, draw it
                action = self._draw_annotation_line(f'{name} unmatched!', len(annotations_time)-1, annotations_time[-1], annotations_frame[-1], math.ceil(len(annotations_time)/2), size.y, border_color)
                if action=='delete_timepoint':
                    self._request_delete(event, annotations_frame[-1])
        else:
            # points in time: draw as lines
            for i,(at,af) in enumerate(zip(annotations_time,annotations_frame)):
                action = self._draw_annotation_line(name, i, at, af, i+1, size.y, border_color)
                if action=='delete_timepoint':
                    self._request_delete(event, annotations_frame[event][i])

        # draw track header (name + background)
        if self._show_annotation_labels:
            draw_list.add_rect_filled(
                (cursor_pos.x+self._h_scroll, cursor_pos.y),
                (cursor_pos.x+self._h_scroll+clip_title_max.x, cursor_pos.y+clip_title_max.y),
                imgui.color_convert_float4_to_u32(color.value)
            )
            draw_list.add_text(
                (cursor_pos.x+text_offset.x+self._h_scroll, cursor_pos.y+text_offset.y),
                text_color_adjusted,
                name
            )
            if event in self._annotation_tooltips and self._annotation_tooltips[event]:
                cpos = imgui.get_cursor_pos()
                imgui.set_cursor_pos((cpos.x+self._h_scroll, cpos.y))
                imgui.invisible_button(f'##track_annotation_label_{name}',clip_title_max)
                if imgui.is_item_hovered():
                    imgui.begin_tooltip()
                    imgui.text(self._annotation_tooltips[event])
                    imgui.end_tooltip()
                imgui.set_cursor_pos(cpos)

        draw_list.pop_clip_rect()
        imgui.dummy(size)   # occupy the space of the track so cursor position moves correctly

    def draw(self):
        self.content_origin = imgui.get_cursor_screen_pos()
        flags = imgui.WindowFlags_.no_background
        if self._allow_timeline_zoom:
            flags |= imgui.WindowFlags_.always_horizontal_scrollbar
        available = imgui.get_content_region_avail()
        draw_width = available.x*self._scale
        if draw_width != self.draw_width:
            self.draw_width = draw_width
            self._calc_scale_fac()
            self._determine_ticks()
        # deal with scroll requests: submit content size and scroll for next window to avoid glitch (one frame delay in scroll)
        imgui.set_next_window_content_size((self.draw_width,0))
        if self._new_h_scroll is not None:
            imgui.set_next_window_scroll((self._new_h_scroll, -1))
            self._new_h_scroll = None
        imgui.begin_child('##all', size=available, window_flags=flags)
        self._h_scroll = imgui.get_scroll_x()
        imgui.begin_child('##timeline',size=(self.draw_width,0),window_flags=imgui.WindowFlags_.no_background)
        imgui.push_style_var(imgui.StyleVar_.item_spacing, (0., 0.))
        self._draw_time_ruler()
        imgui.pop_style_var()

        # draw tracks on the timeline (if any)
        if self._annotations_frame:
            self._draw_tracks()
        imgui.end_child()
        imgui.end_child()


import numpy as np
import pandas as pd
import datetime
import bisect
import pathlib
import enum

from . import utils

class Timestamp:
    def __init__(self, unix_time: int | float, format="%Y-%m-%d %H:%M:%S"):
        self.format = format
        self.display = ""
        self.value = 0
        self.update(unix_time)

    def update(self, unix_time: int | float):
        self.value = int(unix_time)
        if self.value == 0:
            self.display = ""
        else:
            self.display = datetime.datetime.fromtimestamp(unix_time).strftime(self.format)

utils.register_type(utils.CustomTypeEntry(Timestamp,'__Timestamp__',lambda x: x.value, lambda x: Timestamp(x)))


# for reading video timestamp files
class Type(enum.Enum):
    Normal      = enum.auto()
    Stretched   = enum.auto()


class VideoTimestamps:
    def __init__(self, fileName: str|pathlib.Path):
        self.timestamp_dict : dict[int,float] = {}
        self.indices        : list[int] = []
        self.timestamps     : list[float] = []
        self._ifi           : float = None

        df = pd.read_csv(fileName, delimiter='\t', index_col='frame_idx')
        self.timestamp_dict = df.to_dict()['timestamp']

        df = df.reset_index()
        df = df[df['frame_idx']!=-1]
        self.indices    = df['frame_idx'].to_list()
        self.timestamps = df['timestamp'].to_list()

        self.timestamp_stretched_dict   : dict[int,float] = None
        self.timestamps_stretched       : list[float] = None
        self._ifi_stretched             : float = None
        self.has_stretched = 'timestamp_stretched' in df.columns
        if self.has_stretched:
            self.timestamp_stretched_dict = df.to_dict()['timestamp_stretched']
            self.timestamps_stretched = df['timestamp_stretched'].to_list()


    def get_timestamp(self, idx: int, which: Type=Type.Normal) -> float:
        idx = int(idx)
        match which:
            case Type.Normal:
                d = self.timestamp_dict
            case Type.Stretched:
                if not self.has_stretched:
                    raise RuntimeError('stretched timestamps are not available for this video')
                d = self.timestamp_stretched_dict
        if idx in d:
            return d[idx]
        else:
            return -1.

    def find_frame(self, ts: float, which: Type=Type.Normal) -> int:
        match which:
            case Type.Normal:
                timestamps = self.timestamps
            case Type.Stretched:
                if not self.has_stretched:
                    raise RuntimeError('stretched timestamps are not available for this video')
                timestamps = self.timestamps_stretched

        idx = bisect.bisect(timestamps, ts)
        # return nearest
        if idx>=len(timestamps):
            return self.indices[-1]
        elif idx>0 and abs(timestamps[idx-1]-ts)<abs(timestamps[idx]-ts):
            return self.indices[idx-1]
        else:
            return self.indices[idx]

    def get_last(self, which: Type=Type.Normal) -> tuple[int,float]:
        match which:
            case Type.Normal:
                timestamps = self.timestamps
            case Type.Stretched:
                if not self.has_stretched:
                    raise RuntimeError('stretched timestamps are not available for this video')
                timestamps = self.timestamps_stretched
        return self.indices[-1], timestamps[-1]

    def get_IFI(self, which: Type=Type.Normal) -> float:
        match which:
            case Type.Normal:
                if self._ifi is None:
                    self._ifi = np.mean(np.diff(self.timestamps))
                return self._ifi
            case Type.Stretched:
                if not self.has_stretched:
                    raise RuntimeError('stretched timestamps are not available for this video')
                if self._ifi_stretched is None:
                    self._ifi_stretched = np.mean(np.diff(self.timestamps_stretched))
                return self._ifi_stretched


import sys
import enum

class Os(enum.Enum):
    Windows = enum.auto()
    MacOS   = enum.auto()
    Linux   = enum.auto()

if sys.platform.startswith("win"):
    os = Os.Windows
elif sys.platform.startswith("linux"):
    os = Os.Linux
elif sys.platform.startswith("darwin"):
    os = Os.MacOS
else:
    print("Your system is not officially supported at the moment!\n"
          "You can let me know on GitHub, or you can try porting yourself")
    sys.exit(1)